"""
Module to provide ceph control with salt.

:depends:   - ceph_cfg Python module

.. versionadded:: 2016.11.0
"""

import logging

log = logging.getLogger(__name__)

__virtualname__ = "ceph"

try:
    import ceph_cfg

    HAS_CEPH_CFG = True
except ImportError:
    HAS_CEPH_CFG = False


def __virtual__():
    if HAS_CEPH_CFG is False:
        msg = "ceph_cfg unavailable: {} execution module cant be loaded ".format(
            __virtualname__
        )
        return False, msg
    return __virtualname__


def partition_list():
    """
    List partitions by disk

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.partition_list
    """
    return ceph_cfg.partition_list()


def partition_list_osd():
    """
    List all OSD data partitions by partition

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.partition_list_osd
    """
    return ceph_cfg.partition_list_osd()


def partition_list_journal():
    """
    List all OSD journal partitions by partition

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.partition_list_journal
    """
    return ceph_cfg.partition_list_journal()


def osd_discover():
    """
    List all OSD by cluster

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.osd_discover

    """
    return ceph_cfg.osd_discover()


def partition_is(dev):
    """
    Check whether a given device path is a partition or a full disk.

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.partition_is /dev/sdc1
    """
    return ceph_cfg.partition_is(dev)


def zap(target=None, **kwargs):
    """
    Destroy the partition table and content of a given disk.

    .. code-block:: bash

        salt '*' ceph.osd_prepare 'dev'='/dev/vdc' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    dev
        The block device to format.

    cluster_name
        The cluster name. Defaults to ``ceph``.

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.
    """
    if target is not None:
        log.warning("Deprecated use of function, use kwargs")
    target = kwargs.get("dev", target)
    kwargs["dev"] = target
    return ceph_cfg.zap(**kwargs)


def osd_prepare(**kwargs):
    """
    Prepare an OSD

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.osd_prepare 'osd_dev'='/dev/vdc' \\
                'journal_dev'='device' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid' \\
                'osd_fs_type'='xfs' \\
                'osd_uuid'='2a143b73-6d85-4389-a9e9-b8a78d9e1e07' \\
                'journal_uuid'='4562a5db-ff6f-4268-811d-12fd4a09ae98'

    cluster_uuid
        The device to store the osd data on.

    journal_dev
        The journal device. defaults to osd_dev.

    cluster_name
        The cluster name. Defaults to ``ceph``.

    cluster_uuid
        The cluster date will be added too. Defaults to the value found in local config.

    osd_fs_type
        set the file system to store OSD data with. Defaults to "xfs".

    osd_uuid
        set the OSD data UUID. If set will return if OSD with data UUID already exists.

    journal_uuid
        set the OSD journal UUID. If set will return if OSD with journal UUID already exists.
    """
    return ceph_cfg.osd_prepare(**kwargs)


def osd_activate(**kwargs):
    """
    Activate an OSD

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.osd_activate 'osd_dev'='/dev/vdc'
    """
    return ceph_cfg.osd_activate(**kwargs)


def keyring_create(**kwargs):
    """
    Create keyring for cluster

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.keyring_create \\
                'keyring_type'='admin' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    keyring_type (required)
        One of ``admin``, ``mon``, ``osd``, ``rgw``, ``mds``

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.keyring_create(**kwargs)


def keyring_save(**kwargs):
    """
    Create save keyring locally

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.keyring_save \\
                'keyring_type'='admin' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    keyring_type (required)
        One of ``admin``, ``mon``, ``osd``, ``rgw``, ``mds``

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.keyring_save(**kwargs)


def keyring_purge(**kwargs):
    """
    Delete keyring for cluster

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.keyring_purge \\
                'keyring_type'='admin' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    keyring_type (required)
        One of ``admin``, ``mon``, ``osd``, ``rgw``, ``mds``

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.

    If no ceph config file is found, this command will fail.
    """
    return ceph_cfg.keyring_purge(**kwargs)


def keyring_present(**kwargs):
    """
    Returns ``True`` if the keyring is present on disk, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.keyring_present \\
                'keyring_type'='admin' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    keyring_type (required)
        One of ``admin``, ``mon``, ``osd``, ``rgw``, ``mds``

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.keyring_present(**kwargs)


def keyring_auth_add(**kwargs):
    """
    Add keyring to authorized list

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.keyring_auth_add \\
                'keyring_type'='admin' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    keyring_type (required)
        One of ``admin``, ``mon``, ``osd``, ``rgw``, ``mds``

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.keyring_auth_add(**kwargs)


def keyring_auth_del(**kwargs):
    """
    Remove keyring from authorised list

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.keyring_osd_auth_del \\
                'keyring_type'='admin' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    keyring_type (required)
        One of ``admin``, ``mon``, ``osd``, ``rgw``, ``mds``

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.keyring_auth_del(**kwargs)


def mon_is(**kwargs):
    """
    Returns ``True`` if the target is a mon node, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.mon_is \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_name
        The cluster name. Defaults to ``ceph``.

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.
    """
    return ceph_cfg.mon_is(**kwargs)


def mon_status(**kwargs):
    """
    Get status from mon daemon

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.mon_status \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.status(**kwargs)


def mon_quorum(**kwargs):
    """
    Returns ``True`` if the mon daemon is in the quorum, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.mon_quorum \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.mon_quorum(**kwargs)


def mon_active(**kwargs):
    """
    Returns ``True`` if the mon daemon is running, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.mon_active \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.mon_active(**kwargs)


def mon_create(**kwargs):
    """
    Create a mon node

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.mon_create \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.mon_create(**kwargs)


def rgw_pools_create(**kwargs):
    """
    Create pools for rgw

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.rgw_pools_create

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.rgw_pools_create(**kwargs)


def rgw_pools_missing(**kwargs):
    """
    Show pools missing for rgw

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.rgw_pools_missing

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.rgw_pools_missing(**kwargs)


def rgw_create(**kwargs):
    """
    Create a rgw

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.rgw_create \\
                'name' = 'rgw.name' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    name (required)
        The RGW client name. Must start with ``rgw.``

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.rgw_create(**kwargs)


def rgw_destroy(**kwargs):
    """
    Remove a rgw

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.rgw_destroy \\
                'name' = 'rgw.name' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    name (required)
        The RGW client name (must start with ``rgw.``)

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.rgw_destroy(**kwargs)


def mds_create(**kwargs):
    """
    Create a mds

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.mds_create \\
                'name' = 'mds.name' \\
                'port' = 1000, \\
                'addr' = 'fqdn.example.org' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    name (required)
        The MDS name (must start with ``mds.``)

    port (required)
        Port to which the MDS will listen

    addr (required)
        Address or IP address for the MDS to listen

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.mds_create(**kwargs)


def mds_destroy(**kwargs):
    """
    Remove a mds

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.mds_destroy \\
                'name' = 'mds.name' \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    name (required)
        The MDS name (must start with ``mds.``)

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.mds_destroy(**kwargs)


def keyring_auth_list(**kwargs):
    """
    List all cephx authorization keys

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.keyring_auth_list \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_name
        The cluster name. Defaults to ``ceph``.

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.
    """
    return ceph_cfg.keyring_auth_list(**kwargs)


def pool_list(**kwargs):
    """
    List all pools

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.pool_list \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_name
        The cluster name. Defaults to ``ceph``.

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.
    """
    return ceph_cfg.pool_list(**kwargs)


def pool_add(pool_name, **kwargs):
    """
    Create a pool

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.pool_add pool_name \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_name
        The cluster name. Defaults to ``ceph``.

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    pg_num
        Default to 8

    pgp_num
        Default to pg_num

    pool_type
        can take values "replicated" or "erasure"

    erasure_code_profile
        The "erasure_code_profile"

    crush_ruleset
        The crush map rule set
    """
    return ceph_cfg.pool_add(pool_name, **kwargs)


def pool_del(pool_name, **kwargs):
    """
    Delete a pool

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.pool_del pool_name \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_name
        The cluster name. Defaults to ``ceph``.

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.
    """
    return ceph_cfg.pool_del(pool_name, **kwargs)


def purge(**kwargs):
    """
    purge ceph configuration on the node

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.purge \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_name
        The cluster name. Defaults to ``ceph``.

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.
    """
    return ceph_cfg.purge(**kwargs)


def ceph_version():
    """
    Get the version of ceph installed

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.ceph_version
    """
    return ceph_cfg.ceph_version()


def cluster_quorum(**kwargs):
    """
    Get the cluster's quorum status

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.cluster_quorum \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.cluster_quorum(**kwargs)


def cluster_status(**kwargs):
    """
    Get the cluster status, including health if in quorum

    CLI Example:

    .. code-block:: bash

        salt '*' ceph.cluster_status \\
                'cluster_name'='ceph' \\
                'cluster_uuid'='cluster_uuid'

    cluster_uuid
        The cluster UUID. Defaults to value found in ceph config file.

    cluster_name
        The cluster name. Defaults to ``ceph``.
    """
    return ceph_cfg.cluster_status(**kwargs)



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    salt.config.schemas.common
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Common salt configuration schemas
"""


from salt.utils.schema import ArrayItem, OneOfItem, Schema, StringItem


class DefaultIncludeConfig(StringItem):
    """
    Per default, the {0}, will automatically include all config files
    from '{1}/*.conf' ('{1}' is a sub-directory in the same directory
    as the main {0} config file).
    """

    __target__ = None
    __confd_directory__ = None

    title = "Include Config"
    description = __doc__

    def __init__(self, default=None, pattern=None, **kwargs):
        default = "{}/*.conf".format(self.__confd_directory__)
        pattern = r"(?:.*)/\*\.conf"
        super().__init__(default=default, pattern=pattern, **kwargs)

    def __validate_attributes__(self):
        self.__doc__ = DefaultIncludeConfig.__doc__.format(
            self.__target__, self.__confd_directory__
        )
        super().__validate_attributes__()

    def __get_description__(self):
        return self.__doc__.format(self.__target__, self.__confd_directory__)


class MinionDefaultInclude(DefaultIncludeConfig):
    __target__ = "minion"
    __confd_directory__ = "minion.d"


class MasterDefaultInclude(DefaultIncludeConfig):
    __target__ = "master"
    __confd_directory = "master.d"


class IncludeConfig(Schema):
    title = "Include Configuration File(s)"
    description = "Include one or more specific configuration files"

    string_or_array = OneOfItem(items=(StringItem(), ArrayItem(items=StringItem())))(
        flatten=True
    )



"""
    tests.pytests.unit.beacons.test_bonjour_announce
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Bonjour announce beacon test cases
"""
import pytest

import salt.beacons.bonjour_announce as bonjour_announce


@pytest.fixture
def configure_loader_modules():
    return {
        bonjour_announce: {"last_state": {}, "last_state_extra": {"no_devices": False}}
    }


def test_non_list_config():
    config = {}

    ret = bonjour_announce.validate(config)
    assert ret == (False, "Configuration for bonjour_announce beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = bonjour_announce.validate(config)
    assert ret == (
        False,
        "Configuration for bonjour_announce beacon must contain servicetype, port and"
        " txt items.",
    )



"""
    :codeauthor: :email:`Gareth J. Greenaway <gareth@saltstack.com>`
    :codeauthor: :email:`David Murphy <dmurphy@saltstack.com>`
"""

import datetime
import logging
import pathlib
import shutil
import subprocess
import time
import types

import psutil
import pytest

import salt.modules.gpg as gpg
from tests.support.mock import MagicMock, call, patch

pytest.importorskip("gnupg")

pytestmark = [
    pytest.mark.skip_unless_on_linux,
    pytest.mark.requires_random_entropy,
    pytest.mark.slow_test,
]

log = logging.getLogger(__name__)


GPG_TEST_KEY_PASSPHRASE = "testkeypassphrase"
GPG_TEST_KEY_ID = "7416F045"
GPG_TEST_PUB_KEY = """-----BEGIN PGP PUBLIC KEY BLOCK-----
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=yF8M
-----END PGP PUBLIC KEY BLOCK-----
"""

GPG_TEST_PRIV_KEY = """-----BEGIN PGP PRIVATE KEY BLOCK-----


OZV2Hg+93dg3Wi6g/JW4OuTKWKuHRqpRB1J4i4lO
=WRTN
-----END PGP PRIVATE KEY BLOCK-----
"""


@pytest.fixture
def gpghome(tmp_path):
    root = tmp_path / "gpghome"
    root.mkdir(mode=0o0700)
    root.joinpath(".gnupg").mkdir(mode=0o0700)
    pub = root / "gpgfile.pub"
    pub.write_text(GPG_TEST_PUB_KEY)
    priv = root / "gpgfile.priv"
    priv.write_text(GPG_TEST_PRIV_KEY)
    try:
        yield types.SimpleNamespace(path=root, pub=pub, priv=priv)
    finally:
        # Make sure we don't leave any gpg-agent's running behind
        gpg_connect_agent = shutil.which("gpg-connect-agent")
        if gpg_connect_agent:
            gnupghome = root / ".gnupg"
            if not gnupghome.is_dir():
                gnupghome = root
            try:
                subprocess.run(
                    [gpg_connect_agent, "killagent", "/bye"],
                    env={"GNUPGHOME": str(gnupghome)},
                    shell=False,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            except subprocess.CalledProcessError:
                # This is likely CentOS 7 or Amazon Linux 2
                pass

        # If the above errored or was not enough, as a last resort, let's check
        # the running processes.
        for proc in psutil.process_iter():
            try:
                if "gpg-agent" in proc.name():
                    for arg in proc.cmdline():
                        if str(root) in arg:
                            proc.terminate()
            except Exception:  # pylint: disable=broad-except
                pass


@pytest.fixture
def configure_loader_modules(gpghome):
    return {gpg: {}}


def test_list_keys():
    """
    Test gpg.list_keys
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": "/root",
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _list_result = [
        {
            "dummy": "",
            "keyid": "xxxxxxxxxxxxxxxx",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                ["xxxxxxxxxxxxxxxx", "e", "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "date": "1506612692",
            "trust": "-",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        },
        {
            "dummy": "",
            "keyid": "yyyyyyyyyyyyyyyy",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                [
                    "yyyyyyyyyyyyyyyy",
                    "e",
                    "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy",
                ]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy",
            "date": "1506612692",
            "trust": "r",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        },
    ]

    _expected_result = [
        {
            "keyid": "xxxxxxxxxxxxxxxx",
            "uids": ["GPG Person <person@example.com>"],
            "created": "2017-09-28",
            "expires": "2033-09-24",
            "keyLength": "4096",
            "ownerTrust": "Unknown",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "trust": "Unknown",
        },
        {
            "keyid": "yyyyyyyyyyyyyyyy",
            "uids": ["GPG Person <person@example.com>"],
            "created": "2017-09-28",
            "expires": "2033-09-24",
            "keyLength": "4096",
            "ownerTrust": "Unknown",
            "fingerprint": "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy",
            "trust": "Revoked",
        },
    ]

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch.object(gpg, "_list_keys", return_value=_list_result):
                assert gpg.list_keys() == _expected_result


def test_get_key():
    """
    Test gpg.get_key
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": "/root",
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _list_result = [
        {
            "dummy": "",
            "keyid": "xxxxxxxxxxxxxxxx",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                ["xxxxxxxxxxxxxxxx", "e", "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "date": "1506612692",
            "trust": "-",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        }
    ]

    _expected_result = {
        "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "keyid": "xxxxxxxxxxxxxxxx",
        "uids": ["GPG Person <person@example.com>"],
        "created": "2017-09-28",
        "trust": "Unknown",
        "ownerTrust": "Unknown",
        "expires": "2033-09-24",
        "keyLength": "4096",
    }

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch.object(gpg, "_list_keys", return_value=_list_result):
                ret = gpg.get_key("xxxxxxxxxxxxxxxx")
                assert ret == _expected_result


def test_delete_key_without_passphrase(gpghome):
    """
    Test gpg.delete_key without passphrase
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _list_result = [
        {
            "dummy": "",
            "keyid": "xxxxxxxxxxxxxxxx",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                ["xxxxxxxxxxxxxxxx", "e", "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "date": "1506612692",
            "trust": "-",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        }
    ]

    _expected_result = {
        "res": True,
        "message": (
            "Secret key for xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx deleted\nPublic"
            " key for xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx deleted"
        ),
    }

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch.object(gpg, "_list_keys", return_value=_list_result):
                with patch(
                    "salt.modules.gpg.gnupg.GPG.delete_keys",
                    MagicMock(return_value="ok"),
                ) as gnupg_delete_keys:
                    ret = gpg.delete_key(
                        "xxxxxxxxxxxxxxxx", delete_secret=True, use_passphrase=False
                    )
                    assert ret == _expected_result
                    gnupg_delete_keys.assert_called_with(
                        "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
                        False,
                        expect_passphrase=False,
                    )


def test_delete_key_with_passphrase_without_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.delete_key with passphrase but without gpg_passphrase pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _list_result = [
        {
            "dummy": "",
            "keyid": "xxxxxxxxxxxxxxxx",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                ["xxxxxxxxxxxxxxxx", "e", "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "date": "1506612692",
            "trust": "-",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        }
    ]

    _expected_result = {
        "res": True,
        "message": "gpg_passphrase not available in pillar.",
    }

    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=None)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            with patch.object(gpg, "_list_keys", return_value=_list_result):
                with patch(
                    "salt.modules.gpg.gnupg.GPG.delete_keys"
                ) as gnupg_delete_keys:
                    ret = gpg.delete_key("xxxxxxxxxxxxxxxx", delete_secret=True)
                    assert ret == _expected_result
                    gnupg_delete_keys.assert_not_called()


def test_delete_key_with_passphrase_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.delete_key with passphrase and gpg_passphrase pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _list_result = [
        {
            "dummy": "",
            "keyid": "xxxxxxxxxxxxxxxx",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                ["xxxxxxxxxxxxxxxx", "e", "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "date": "1506612692",
            "trust": "-",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        }
    ]

    _expected_result = {
        "res": True,
        "message": (
            "Secret key for xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx deleted\nPublic"
            " key for xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx deleted"
        ),
    }

    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            with patch.object(gpg, "_list_keys", return_value=_list_result):
                with patch(
                    "salt.modules.gpg.gnupg.GPG.delete_keys",
                    MagicMock(return_value="ok"),
                ) as gnupg_delete_keys:
                    ret = gpg.delete_key("xxxxxxxxxxxxxxxx", delete_secret=True)
                    assert ret == _expected_result
                    gnupg_delete_keys.assert_called_with(
                        "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
                        False,
                        passphrase=GPG_TEST_KEY_PASSPHRASE,
                    )


def test_export_public_key(gpghome):
    """
    Test gpg.export_key with single public key
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch(
                "salt.modules.gpg.gnupg.GPG.export_keys",
                MagicMock(return_value=GPG_TEST_PUB_KEY),
            ) as gnupg_export_keys:
                ret = gpg.export_key("xxxxxxxxxxxxxxxx", bare=True)
                assert ret == GPG_TEST_PUB_KEY
                gnupg_export_keys.assert_called_with(
                    ["xxxxxxxxxxxxxxxx"],
                    False,
                    expect_passphrase=False,
                )


def test_export_public_key_to_file(gpghome):
    """
    Test gpg.export_key output public key to file
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    exported_keyfile = gpghome.path / "exported_pub_key"
    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)

    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch(
                "salt.modules.gpg.gnupg.GPG.export_keys",
                MagicMock(return_value=GPG_TEST_PUB_KEY),
            ) as gnupg_export_keys:
                ret = gpg.export_key(
                    keyids="xxxxxxxxxxxxxxxx", output=exported_keyfile, bare=True
                )
                assert ret == GPG_TEST_PUB_KEY
                keyfile_contents = pathlib.Path(exported_keyfile).read_text()
                assert keyfile_contents == GPG_TEST_PUB_KEY


def test_export_multiple_public_keys(gpghome):
    """
    Test gpg.export_key with multiple public keys
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch(
                "salt.modules.gpg.gnupg.GPG.export_keys",
                MagicMock(return_value=GPG_TEST_PUB_KEY),
            ) as gnupg_export_keys:
                ret = gpg.export_key(
                    "xxxxxxxxxxxxxxxx,yyyyyyyyyyyyyyyy,zzzzzzzzzzzzzzzz",
                    bare=True,
                )
                assert ret == GPG_TEST_PUB_KEY
                gnupg_export_keys.assert_called_with(
                    ["xxxxxxxxxxxxxxxx", "yyyyyyyyyyyyyyyy", "zzzzzzzzzzzzzzzz"],
                    False,
                    expect_passphrase=False,
                )


def test_export_secret_key_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.export_key with secret key and gpg_passphrase in pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            with patch(
                "salt.modules.gpg.gnupg.GPG.export_keys",
                MagicMock(return_value=GPG_TEST_PRIV_KEY),
            ) as gnupg_export_keys:
                ret = gpg.export_key(
                    "xxxxxxxxxxxxxxxx", secret=True, use_passphrase=True, bare=True
                )
                assert ret == GPG_TEST_PRIV_KEY
                gnupg_export_keys.assert_called_with(
                    ["xxxxxxxxxxxxxxxx"],
                    True,
                    passphrase=GPG_TEST_KEY_PASSPHRASE,
                )


def test_export_secret_key_to_file_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.export_key output secret key to file with gpg_passphrase in pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    exported_keyfile = gpghome.path / "exported_priv_key"
    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            with patch(
                "salt.modules.gpg.gnupg.GPG.export_keys",
                MagicMock(return_value=GPG_TEST_PRIV_KEY),
            ) as gnupg_export_keys:
                ret = gpg.export_key(
                    keyids="xxxxxxxxxxxxxxxx",
                    secret=True,
                    output=exported_keyfile,
                    use_passphrase=True,
                    bare=True,
                )
                assert ret == GPG_TEST_PRIV_KEY
                gnupg_export_keys.assert_called_with(
                    ["xxxxxxxxxxxxxxxx"],
                    True,
                    passphrase=GPG_TEST_KEY_PASSPHRASE,
                )
                keyfile_contents = pathlib.Path(exported_keyfile).read_text()
                assert keyfile_contents == GPG_TEST_PRIV_KEY


def test_create_key_without_passphrase(gpghome):
    """
    Test gpg.create_key without a passphrase
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            ret = gpg.create_key()
            assert ret["res"]
            assert ret["fingerprint"]
            assert ret["message"] == "GPG key pair successfully generated."


def test_create_key_with_passphrase_without_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.create_key with passphrase but without gpg_passphrase pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=None)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            ret = gpg.create_key(use_passphrase=True)
            assert not ret["res"]
            assert ret["message"] == "gpg_passphrase not available in pillar."


def test_create_key_with_passphrase_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.create_key with passphrase and gpg_passphrase pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            ret = gpg.create_key(use_passphrase=True)
            assert ret["res"]
            assert ret["fingerprint"]
            assert ret["message"] == "GPG key pair successfully generated."


def test_search_keys(gpghome):
    """
    Test gpg.search_keys
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _search_result = [
        {
            "keyid": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "uids": ["GPG Person <person@example.com>"],
            "expires": "",
            "sigs": [],
            "length": "1024",
            "algo": "17",
            "date": int(time.mktime(datetime.datetime(2004, 11, 13).timetuple())),
            "type": "pub",
        }
    ]

    _expected_result = [
        {
            "uids": ["GPG Person <person@example.com>"],
            "created": "2004-11-13",
            "keyLength": "1024",
            "keyid": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        }
    ]

    mock_search_keys = MagicMock(return_value=_search_result)
    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch.object(gpg, "_search_keys", mock_search_keys):
                ret = gpg.search_keys("person@example.com")
                assert ret == _expected_result

                assert (
                    call("person@example.com", "keys.openpgp.org", None)
                    in mock_search_keys.mock_calls
                )

                ret = gpg.search_keys("person@example.com", "keyserver.ubuntu.com")
                assert ret == _expected_result

                assert (
                    call("person@example.com", "keyserver.ubuntu.com", None)
                    in mock_search_keys.mock_calls
                )


def test_gpg_import_pub_key(gpghome):
    config_user = MagicMock(return_value="salt")
    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            ret = gpg.import_key(None, str(gpghome.pub), "salt", str(gpghome.path))
            assert ret["res"] is True


def test_gpg_import_priv_key(gpghome):
    config_user = MagicMock(return_value="salt")
    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            ret = gpg.import_key(None, str(gpghome.priv), "salt", str(gpghome.path))
            assert ret["res"] is True


def test_gpg_sign(gpghome):
    """
    Test gpg.sign
    """

    config_user = MagicMock(return_value="salt")
    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            with patch.dict(gpg.__salt__, {"pillar.get": pillar_mock}):
                key = gpg.import_key(None, str(gpghome.priv), "salt", str(gpghome.path))
                assert key["res"] is True
                gpg_text_input = "The quick brown fox jumped over the lazy dog"
                ret = gpg.sign(
                    keyid=GPG_TEST_KEY_ID,
                    text=gpg_text_input,
                    use_passphrase=True,
                    gnupghome=str(gpghome.path),
                )
                assert "-----BEGIN PGP SIGNATURE-----" in str(ret)


def test_gpg_encrypt_message(gpghome):
    """
    Test gpg.encrypt
    """

    config_user = MagicMock(return_value="salt")

    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            key = gpg.import_key(None, str(gpghome.priv), "salt", str(gpghome.path))
            assert key["res"] is True
            gpg_text_input = "The quick brown fox jumped over the lazy dog"
            ret = gpg.encrypt(
                recipients=GPG_TEST_KEY_ID,
                text=gpg_text_input,
                use_passphrase=True,
                always_trust=True,
                gnupghome=str(gpghome.path),
            )
            assert ret["res"] is True


def test_gpg_encrypt_and_sign_message_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.encrypt sign message with passphrase and gpg_passphrase in pillar
    """

    config_user = MagicMock(return_value="salt")

    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            with patch.dict(gpg.__salt__, {"pillar.get": pillar_mock}):
                key = gpg.import_key(None, str(gpghome.priv), "salt", str(gpghome.path))
                assert key["res"] is True
                gpg_text_input = "The quick brown fox jumped over the lazy dog"
                ret = gpg.encrypt(
                    recipients=GPG_TEST_KEY_ID,
                    text=gpg_text_input,
                    sign=True,
                    use_passphrase=True,
                    always_trust=True,
                    gnupghome=str(gpghome.path),
                )
                assert ret["res"] is True


def test_gpg_decrypt_message_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.decrypt with passphrase and gpg_passphrase in pillar
    """

    gpg_encrypted_message = """-----BEGIN PGP MESSAGE-----
    hQGMA7z9rKs9ZvTOAQwAnMbwchCm1VXOD+Ml0rnNrhDhsRm+6O96FOq5lWY0ntkj
    vnXeFOgUf0wzK4hkQT/Yo4/ZpDkV3iwwSIjesqNDS1U/KWfbe2pFeph6w9fHFnXf
    e8RJY8OGJHN8A9TlCIyKDVKGoXZEcSDZ4K0fNL/OxF3LxsSjJ894F+e2MCnoQZgq
    3uxQFYb9Gec8iPpBcd7s74J8cX1To9rBKRNSP73NRwVvKbPgB0NjzrLVPmqHGO7Q
    wFf9tZdGYmxV/AB9H0gCUITBCwnndissVqFpjUpVEIL1jOBus73oiL9Grn5QlZic
    wDOZyKDlWR15jiiDaWvq4ES6O0j2Oj/FodItJhTbzeXEFOxn7cNbW2kAAc08Hs1u
    zPI2LywaCSM/IOxrRI8n/ExXxtU0k48827AWTLYmhm+vDct7gQHEJ8Qu3/d5FsP/
    f1qWrtl+E7TpIIlUIgtPL596LYiwQLisqh1JeqqleSnBESIzgg7teolYdYED606z
    j9a1oTUHvSnKz6Q5Yh850mYBCBc/iEn0gJBgseGK6rXJF54f5duybkuTKrV+7hUd
    euy/+K0mnQWPS0G9NhCdQAaydNBZywOyQU6Yv1gPHapRrHOkhZq51sBKmZRGdqH/
    llX+Wd7pUetBV+vmZp75l8q9CNVpefA=
    =WXYF
    -----END PGP MESSAGE-----
    """

    config_user = MagicMock(return_value="salt")
    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            with patch.dict(gpg.__salt__, {"pillar.get": pillar_mock}):
                key = gpg.import_key(None, str(gpghome.priv), "salt", str(gpghome.path))
                assert key["res"] is True
                ret = gpg.decrypt(
                    text=gpg_encrypted_message,
                    use_passphrase=True,
                    gnupghome=str(gpghome.path),
                )
                assert ret["res"] is True



"""
Unit tests for the Default Job Cache (local_cache).
"""

import os
import time

import pytest

import salt.returners.local_cache as local_cache
import salt.utils.files
import salt.utils.jid
import salt.utils.job
import salt.utils.platform
from tests.support.mock import MagicMock, patch


@pytest.fixture
def tmp_cache_dir(tmp_path):
    return tmp_path / "cache_dir"


@pytest.fixture
def tmp_jid_dir(tmp_cache_dir):
    return tmp_cache_dir / "jobs"


@pytest.fixture
def configure_loader_modules(tmp_cache_dir):
    return {
        local_cache: {
            "__opts__": {
                "cachedir": str(tmp_cache_dir),
                "keep_jobs_seconds": 3600,
            },
        }
    }


@pytest.fixture
def make_tmp_jid_dirs(tmp_jid_dir):
    def _make_tmp_jid_dirs(create_files=True):
        """
        Helper function to set up temporary directories and files used for
        testing the clean_old_jobs function.

        This emulates salt.utils.jid.jid_dir() by creating this structure:

        RUNTIME_VARS.TMP_JID_DIR dir/
            random dir from tempfile.mkdtemp/
            'jid' directory/
                'jid' file

        Returns a temp_dir name and a jid_file_path. If create_files is False,
        the jid_file_path will be None.
        """
        # First, create the /tmp/salt_test_job_cache/jobs/ directory to hold jid dirs
        tmp_jid_dir.mkdir(parents=True, exist_ok=True)

        # Then create a JID temp file in "/tmp/salt_test_job_cache/"
        temp_dir = tmp_jid_dir / "tmp_dir"
        temp_dir.mkdir(parents=True, exist_ok=True)

        jid_file_path = None
        if create_files:
            dir_name = temp_dir / "jid"
            dir_name.mkdir(parents=True, exist_ok=True)
            jid_file_path = dir_name / "jid"
            jid_file_path.write_text("this is a jid file")
            jid_file_path = str(jid_file_path)

        return str(temp_dir), jid_file_path

    return _make_tmp_jid_dirs


def test_clean_old_jobs_no_jid_root():
    """
    Tests that the function returns None when no jid_root is found.
    """
    with patch("os.path.exists", MagicMock(return_value=False)):
        assert local_cache.clean_old_jobs() is None


def test_clean_old_jobs_empty_jid_dir_removed(make_tmp_jid_dirs, tmp_jid_dir):
    """
    Tests that an empty JID dir is removed when it is old enough to be deleted.
    """
    # Create temp job cache dir without files in it.
    jid_dir, jid_file = make_tmp_jid_dirs(create_files=False)

    # File timestamps on Windows aren't as precise. Let a little time pass
    if salt.utils.platform.is_windows():
        time.sleep(0.01)

    # Make sure there are no files in the directory before continuing
    assert jid_file is None

    # Call clean_old_jobs function, patching the keep_jobs_seconds value
    # with a very small value to force the call to clean the job.
    with patch.dict(local_cache.__opts__, {"keep_jobs_seconds": 0.00000001}):
        # Sleep on Windows because time.time is only precise to 3 decimal
        # points, and therefore subtracting the jid_ctime from time.time
        # will result in a negative number
        if salt.utils.platform.is_windows():
            time.sleep(0.25)
        local_cache.clean_old_jobs()

    # Assert that the JID dir was removed
    assert [] == os.listdir(tmp_jid_dir)


def test_clean_old_jobs_empty_jid_dir_remains(make_tmp_jid_dirs, tmp_jid_dir):
    """
    Tests that an empty JID dir is NOT removed because it was created within
    the keep_jobs_seconds time frame.
    """
    # Create temp job cache dir without files in it.
    jid_dir, jid_file = make_tmp_jid_dirs(create_files=False)

    # Make sure there are no files in the directory
    assert jid_file is None

    # Call clean_old_jobs function
    local_cache.clean_old_jobs()

    # Get the name of the JID directory that was created to test against
    if salt.utils.platform.is_windows():
        jid_dir_name = jid_dir.rpartition("\\")[2]
    else:
        jid_dir_name = jid_dir.rpartition("/")[2]

    # Assert the JID directory is still present to be cleaned after keep_jobs_seconds interval
    assert [jid_dir_name] == os.listdir(tmp_jid_dir)


def test_clean_old_jobs_jid_file_corrupted(make_tmp_jid_dirs, tmp_jid_dir):
    """
    Tests that the entire JID dir is removed when the jid_file is not a file.
    This scenario indicates a corrupted cache entry, so the entire dir is scrubbed.
    """
    # Create temp job cache dir and jid file
    jid_dir, jid_file = make_tmp_jid_dirs()

    # Make sure there is a jid file in a new job cache director
    jid_dir_name = jid_file.rpartition(os.sep)[2]
    assert jid_dir_name == "jid"

    # Even though we created a valid jid file in the _make_tmp_jid_dirs call to get
    # into the correct loop, we need to mock the 'os.path.isfile' check to force the
    # "corrupted file" check in the clean_old_jobs call.
    with patch("os.path.isfile", MagicMock(return_value=False)) as mock:
        local_cache.clean_old_jobs()

    # there should be only 1 dir in TMP_JID_DIR
    assert 1 == len(os.listdir(tmp_jid_dir))
    # top level dir should still be present
    assert os.path.exists(jid_dir) is True
    assert os.path.isdir(jid_dir) is True
    # while the 'jid' dir inside it should be gone
    assert os.path.exists(jid_dir_name) is False


def test_clean_old_jobs_jid_file_is_cleaned(make_tmp_jid_dirs, tmp_jid_dir):
    """
    Test that the entire JID dir is removed when a job is old enough to be removed.
    """
    # Create temp job cache dir and jid file
    jid_dir, jid_file = make_tmp_jid_dirs()

    # File timestamps on Windows aren't as precise. Let a little time pass
    if salt.utils.platform.is_windows():
        time.sleep(0.01)

    # Make sure there is a jid directory
    jid_dir_name = jid_file.rpartition(os.sep)[2]
    assert jid_dir_name == "jid"

    # Call clean_old_jobs function, patching the keep_jobs_seconds value with a
    # very small value to force the call to clean the job.
    with patch.dict(local_cache.__opts__, {"keep_jobs_seconds": 0.00000001}):
        # Sleep on Windows because time.time is only precise to 3 decimal
        # points, and therefore subtracting the jid_ctime from time.time
        # will result in a negative number
        if salt.utils.platform.is_windows():
            time.sleep(0.25)
        local_cache.clean_old_jobs()

    # there should be only 1 dir in TMP_JID_DIR
    assert 1 == len(os.listdir(tmp_jid_dir))
    # top level dir should still be present
    assert os.path.exists(jid_dir) is True
    assert os.path.isdir(jid_dir) is True
    # while the 'jid' dir inside it should be gone
    assert os.path.exists(jid_dir_name) is False



"""
Nova class
"""


import inspect
import logging
import time

import salt.utils.cloud
import salt.utils.files
from salt.exceptions import SaltCloudSystemExit
from salt.utils.versions import Version

HAS_NOVA = False
# pylint: disable=import-error
try:
    import novaclient
    import novaclient.auth_plugin
    import novaclient.base
    import novaclient.exceptions
    import novaclient.extension
    import novaclient.utils
    from novaclient import client
    from novaclient.shell import OpenStackComputeShell

    HAS_NOVA = True
except ImportError:
    pass

HAS_KEYSTONEAUTH = False
try:
    import keystoneauth1.loading
    import keystoneauth1.session

    HAS_KEYSTONEAUTH = True
except ImportError:
    pass
# pylint: enable=import-error


# Get logging started
log = logging.getLogger(__name__)

# Version added to novaclient.client.Client function
NOVACLIENT_MINVER = "2.6.1"
NOVACLIENT_MAXVER = "6.0.1"

# dict for block_device_mapping_v2
CLIENT_BDM2_KEYS = {
    "id": "uuid",
    "source": "source_type",
    "dest": "destination_type",
    "bus": "disk_bus",
    "device": "device_name",
    "size": "volume_size",
    "format": "guest_format",
    "bootindex": "boot_index",
    "type": "device_type",
    "shutdown": "delete_on_termination",
}


def check_nova():
    if HAS_NOVA:
        novaclient_ver = Version(novaclient.__version__)
        min_ver = Version(NOVACLIENT_MINVER)
        max_ver = Version(NOVACLIENT_MAXVER)
        if min_ver <= novaclient_ver <= max_ver:
            return HAS_NOVA
        elif novaclient_ver > max_ver:
            log.debug(
                "Older novaclient version required. Maximum: %s", NOVACLIENT_MAXVER
            )
            return False
        log.debug("Newer novaclient version required.  Minimum: %s", NOVACLIENT_MINVER)
    return False


if check_nova():
    try:
        import novaclient.auth_plugin
    except ImportError:
        log.debug(
            "Using novaclient version 7.0.0 or newer. Authentication "
            "plugin auth_plugin.py is not available anymore."
        )


# kwargs has to be an object instead of a dictionary for the __post_parse_arg__
class KwargsStruct:
    def __init__(self, **entries):
        self.__dict__.update(entries)


def _parse_block_device_mapping_v2(
    block_device=None, boot_volume=None, snapshot=None, ephemeral=None, swap=None
):
    bdm = []
    if block_device is None:
        block_device = []
    if ephemeral is None:
        ephemeral = []

    if boot_volume is not None:
        bdm_dict = {
            "uuid": boot_volume,
            "source_type": "volume",
            "destination_type": "volume",
            "boot_index": 0,
            "delete_on_termination": False,
        }
        bdm.append(bdm_dict)

    if snapshot is not None:
        bdm_dict = {
            "uuid": snapshot,
            "source_type": "snapshot",
            "destination_type": "volume",
            "boot_index": 0,
            "delete_on_termination": False,
        }
        bdm.append(bdm_dict)

    for device_spec in block_device:
        bdm_dict = {}

        for key, value in device_spec.items():
            bdm_dict[CLIENT_BDM2_KEYS[key]] = value

        # Convert the delete_on_termination to a boolean or set it to true by
        # default for local block devices when not specified.
        if "delete_on_termination" in bdm_dict:
            action = bdm_dict["delete_on_termination"]
            bdm_dict["delete_on_termination"] = action == "remove"
        elif bdm_dict.get("destination_type") == "local":
            bdm_dict["delete_on_termination"] = True

        bdm.append(bdm_dict)

    for ephemeral_spec in ephemeral:
        bdm_dict = {
            "source_type": "blank",
            "destination_type": "local",
            "boot_index": -1,
            "delete_on_termination": True,
        }
        if "size" in ephemeral_spec:
            bdm_dict["volume_size"] = ephemeral_spec["size"]
        if "format" in ephemeral_spec:
            bdm_dict["guest_format"] = ephemeral_spec["format"]

        bdm.append(bdm_dict)

    if swap is not None:
        bdm_dict = {
            "source_type": "blank",
            "destination_type": "local",
            "boot_index": -1,
            "delete_on_termination": True,
            "guest_format": "swap",
            "volume_size": swap,
        }
        bdm.append(bdm_dict)

    return bdm


class NovaServer:
    def __init__(self, name, server, password=None):
        """
        Make output look like libcloud output for consistency
        """
        self.name = name
        self.id = server["id"]
        self.image = server.get("image", {}).get("id", "Boot From Volume")
        self.size = server["flavor"]["id"]
        self.state = server["state"]
        self._uuid = None
        self.extra = {"metadata": server["metadata"], "access_ip": server["accessIPv4"]}

        self.addresses = server.get("addresses", {})
        self.public_ips, self.private_ips = [], []
        self.fixed_ips, self.floating_ips = [], []
        for network in self.addresses.values():
            for addr in network:
                if salt.utils.cloud.is_public_ip(addr["addr"]):
                    self.public_ips.append(addr["addr"])
                else:
                    self.private_ips.append(addr["addr"])
                if addr.get("OS-EXT-IPS:type") == "floating":
                    self.floating_ips.append(addr["addr"])
                else:
                    self.fixed_ips.append(addr["addr"])

        if password:
            self.extra["password"] = password

    def __str__(self):
        return self.__dict__


def get_entry(dict_, key, value, raise_error=True):
    for entry in dict_:
        if entry[key] == value:
            return entry
    if raise_error is True:
        raise SaltCloudSystemExit("Unable to find {} in {}.".format(key, dict_))
    return {}


def get_entry_multi(dict_, pairs, raise_error=True):
    for entry in dict_:
        if all([entry[key] == value for key, value in pairs]):
            return entry
    if raise_error is True:
        raise SaltCloudSystemExit("Unable to find {} in {}.".format(pairs, dict_))
    return {}


def get_endpoint_url_v3(catalog, service_type, region_name):
    for service_entry in catalog:
        if service_entry["type"] == service_type:
            for endpoint_entry in service_entry["endpoints"]:
                if (
                    endpoint_entry["region"] == region_name
                    and endpoint_entry["interface"] == "public"
                ):
                    return endpoint_entry["url"]
    return None


def sanatize_novaclient(kwargs):
    variables = (
        "username",
        "api_key",
        "project_id",
        "auth_url",
        "insecure",
        "timeout",
        "proxy_tenant_id",
        "proxy_token",
        "region_name",
        "endpoint_type",
        "extensions",
        "service_type",
        "service_name",
        "volume_service_name",
        "timings",
        "bypass_url",
        "os_cache",
        "no_cache",
        "http_log_debug",
        "auth_system",
        "auth_plugin",
        "auth_token",
        "cacert",
        "tenant_id",
    )
    ret = {}
    for var in kwargs:
        if var in variables:
            ret[var] = kwargs[var]

    return ret


# Function alias to not shadow built-ins
class SaltNova:
    """
    Class for all novaclient functions
    """

    extensions = []

    def __init__(
        self,
        username,
        project_id,
        auth_url,
        region_name=None,
        password=None,
        os_auth_plugin=None,
        use_keystoneauth=False,
        **kwargs
    ):
        """
        Set up nova credentials
        """
        if all([use_keystoneauth, HAS_KEYSTONEAUTH]):
            self._new_init(
                username=username,
                project_id=project_id,
                auth_url=auth_url,
                region_name=region_name,
                password=password,
                os_auth_plugin=os_auth_plugin,
                **kwargs
            )
        else:
            self._old_init(
                username=username,
                project_id=project_id,
                auth_url=auth_url,
                region_name=region_name,
                password=password,
                os_auth_plugin=os_auth_plugin,
                **kwargs
            )

    def _new_init(
        self,
        username,
        project_id,
        auth_url,
        region_name,
        password,
        os_auth_plugin,
        auth=None,
        verify=True,
        **kwargs
    ):
        if auth is None:
            auth = {}

        loader = keystoneauth1.loading.get_plugin_loader(os_auth_plugin or "password")

        self.client_kwargs = kwargs.copy()
        self.kwargs = auth.copy()
        if not self.extensions:
            if hasattr(OpenStackComputeShell, "_discover_extensions"):
                self.extensions = OpenStackComputeShell()._discover_extensions("2.0")
            else:
                self.extensions = client.discover_extensions("2.0")
            for extension in self.extensions:
                extension.run_hooks("__pre_parse_args__")
            self.client_kwargs["extensions"] = self.extensions

        self.kwargs["username"] = username
        self.kwargs["project_name"] = project_id
        self.kwargs["auth_url"] = auth_url
        self.kwargs["password"] = password
        if auth_url.endswith("3"):
            self.kwargs["user_domain_name"] = kwargs.get("user_domain_name", "default")
            self.kwargs["project_domain_name"] = kwargs.get(
                "project_domain_name", "default"
            )

        self.client_kwargs["region_name"] = region_name
        self.client_kwargs["service_type"] = "compute"

        if hasattr(self, "extensions"):
            # needs an object, not a dictionary
            self.kwargstruct = KwargsStruct(**self.client_kwargs)
            for extension in self.extensions:
                extension.run_hooks("__post_parse_args__", self.kwargstruct)
            self.client_kwargs = self.kwargstruct.__dict__

        # Requires novaclient version >= 2.6.1
        self.version = str(kwargs.get("version", 2))

        self.client_kwargs = sanatize_novaclient(self.client_kwargs)
        options = loader.load_from_options(**self.kwargs)
        self.session = keystoneauth1.session.Session(auth=options, verify=verify)
        conn = client.Client(
            version=self.version, session=self.session, **self.client_kwargs
        )
        self.kwargs["auth_token"] = conn.client.session.get_token()
        identity_service_type = kwargs.get("identity_service_type", "identity")
        self.catalog = (
            conn.client.session.get(
                "/auth/catalog", endpoint_filter={"service_type": identity_service_type}
            )
            .json()
            .get("catalog", [])
        )
        if conn.client.get_endpoint(service_type=identity_service_type).endswith("v3"):
            self._v3_setup(region_name)
        else:
            self._v2_setup(region_name)

    def _old_init(
        self,
        username,
        project_id,
        auth_url,
        region_name,
        password,
        os_auth_plugin,
        **kwargs
    ):
        self.kwargs = kwargs.copy()
        if not self.extensions:
            if hasattr(OpenStackComputeShell, "_discover_extensions"):
                self.extensions = OpenStackComputeShell()._discover_extensions("2.0")
            else:
                self.extensions = client.discover_extensions("2.0")
            for extension in self.extensions:
                extension.run_hooks("__pre_parse_args__")
            self.kwargs["extensions"] = self.extensions

        self.kwargs["username"] = username
        self.kwargs["project_id"] = project_id
        self.kwargs["auth_url"] = auth_url
        self.kwargs["region_name"] = region_name
        self.kwargs["service_type"] = "compute"

        # used in novaclient extensions to see if they are rackspace or not, to know if it needs to load
        # the hooks for that extension or not.  This is cleaned up by sanatize_novaclient
        self.kwargs["os_auth_url"] = auth_url

        if os_auth_plugin is not None:
            novaclient.auth_plugin.discover_auth_systems()
            auth_plugin = novaclient.auth_plugin.load_plugin(os_auth_plugin)
            self.kwargs["auth_plugin"] = auth_plugin
            self.kwargs["auth_system"] = os_auth_plugin

        if not self.kwargs.get("api_key", None):
            self.kwargs["api_key"] = password

        # This has to be run before sanatize_novaclient before extra variables are cleaned out.
        if hasattr(self, "extensions"):
            # needs an object, not a dictionary
            self.kwargstruct = KwargsStruct(**self.kwargs)
            for extension in self.extensions:
                extension.run_hooks("__post_parse_args__", self.kwargstruct)
            self.kwargs = self.kwargstruct.__dict__

        self.kwargs = sanatize_novaclient(self.kwargs)

        # Requires novaclient version >= 2.6.1
        self.kwargs["version"] = str(kwargs.get("version", 2))

        conn = client.Client(**self.kwargs)
        try:
            conn.client.authenticate()
        except novaclient.exceptions.AmbiguousEndpoints:
            raise SaltCloudSystemExit(
                "Nova provider requires a 'region_name' to be specified"
            )

        self.kwargs["auth_token"] = conn.client.auth_token
        self.catalog = conn.client.service_catalog.catalog["access"]["serviceCatalog"]

        self._v2_setup(region_name)

    def _v3_setup(self, region_name):
        if region_name is not None:
            self.client_kwargs["bypass_url"] = get_endpoint_url_v3(
                self.catalog, "compute", region_name
            )
            log.debug("Using Nova bypass_url: %s", self.client_kwargs["bypass_url"])

        self.compute_conn = client.Client(
            version=self.version, session=self.session, **self.client_kwargs
        )

        volume_endpoints = get_entry(
            self.catalog, "type", "volume", raise_error=False
        ).get("endpoints", {})
        if volume_endpoints:
            if region_name is not None:
                self.client_kwargs["bypass_url"] = get_endpoint_url_v3(
                    self.catalog, "volume", region_name
                )
                log.debug(
                    "Using Cinder bypass_url: %s", self.client_kwargs["bypass_url"]
                )

            self.volume_conn = client.Client(
                version=self.version, session=self.session, **self.client_kwargs
            )
            if hasattr(self, "extensions"):
                self.expand_extensions()
        else:
            self.volume_conn = None

    def _v2_setup(self, region_name):
        if region_name is not None:
            servers_endpoints = get_entry(self.catalog, "type", "compute")["endpoints"]
            self.kwargs["bypass_url"] = get_entry(
                servers_endpoints, "region", region_name
            )["publicURL"]

        self.compute_conn = client.Client(**self.kwargs)

        volume_endpoints = get_entry(
            self.catalog, "type", "volume", raise_error=False
        ).get("endpoints", {})
        if volume_endpoints:
            if region_name is not None:
                self.kwargs["bypass_url"] = get_entry(
                    volume_endpoints, "region", region_name
                )["publicURL"]

            self.volume_conn = client.Client(**self.kwargs)
            if hasattr(self, "extensions"):
                self.expand_extensions()
        else:
            self.volume_conn = None

    def expand_extensions(self):
        for connection in (self.compute_conn, self.volume_conn):
            if connection is None:
                continue
            for extension in self.extensions:
                for attr in extension.module.__dict__:
                    if not inspect.isclass(getattr(extension.module, attr)):
                        continue
                    for key, value in connection.__dict__.items():
                        if not isinstance(value, novaclient.base.Manager):
                            continue
                        if value.__class__.__name__ == attr:
                            setattr(
                                connection, key, extension.manager_class(connection)
                            )

    def get_catalog(self):
        """
        Return service catalog
        """
        return self.catalog

    def server_show_libcloud(self, uuid):
        """
        Make output look like libcloud output for consistency
        """
        server_info = self.server_show(uuid)
        server = next(iter(server_info.values()))
        server_name = next(iter(server_info.keys()))
        if not hasattr(self, "password"):
            self.password = None
        ret = NovaServer(server_name, server, self.password)

        return ret

    def boot(self, name, flavor_id=0, image_id=0, timeout=300, **kwargs):
        """
        Boot a cloud server.
        """
        nt_ks = self.compute_conn
        kwargs["name"] = name
        kwargs["flavor"] = flavor_id
        kwargs["image"] = image_id or None
        ephemeral = kwargs.pop("ephemeral", [])
        block_device = kwargs.pop("block_device", [])
        boot_volume = kwargs.pop("boot_volume", None)
        snapshot = kwargs.pop("snapshot", None)
        swap = kwargs.pop("swap", None)
        kwargs["block_device_mapping_v2"] = _parse_block_device_mapping_v2(
            block_device=block_device,
            boot_volume=boot_volume,
            snapshot=snapshot,
            ephemeral=ephemeral,
            swap=swap,
        )
        response = nt_ks.servers.create(**kwargs)
        self.uuid = response.id
        self.password = getattr(response, "adminPass", None)

        start = time.time()
        trycount = 0
        while True:
            trycount += 1
            try:
                return self.server_show_libcloud(self.uuid)
            except Exception as exc:  # pylint: disable=broad-except
                log.debug("Server information not yet available: %s", exc)
                time.sleep(1)
                if time.time() - start > timeout:
                    log.error(
                        "Timed out after %s seconds while waiting for data", timeout
                    )
                    return False

                log.debug("Retrying server_show() (try %s)", trycount)

    def show_instance(self, name):
        """
        Find a server by its name (libcloud)
        """
        return self.server_by_name(name)

    def root_password(self, server_id, password):
        """
        Change server(uuid's) root password
        """
        nt_ks = self.compute_conn
        nt_ks.servers.change_password(server_id, password)

    def server_by_name(self, name):
        """
        Find a server by its name
        """
        return self.server_show_libcloud(self.server_list().get(name, {}).get("id", ""))

    def _volume_get(self, volume_id):
        """
        Organize information about a volume from the volume_id
        """
        if self.volume_conn is None:
            raise SaltCloudSystemExit("No cinder endpoint available")
        nt_ks = self.volume_conn
        volume = nt_ks.volumes.get(volume_id)
        response = {
            "name": volume.display_name,
            "size": volume.size,
            "id": volume.id,
            "description": volume.display_description,
            "attachments": volume.attachments,
            "status": volume.status,
        }
        return response

    def volume_list(self, search_opts=None):
        """
        List all block volumes
        """
        if self.volume_conn is None:
            raise SaltCloudSystemExit("No cinder endpoint available")
        nt_ks = self.volume_conn
        volumes = nt_ks.volumes.list(search_opts=search_opts)
        response = {}
        for volume in volumes:
            response[volume.display_name] = {
                "name": volume.display_name,
                "size": volume.size,
                "id": volume.id,
                "description": volume.display_description,
                "attachments": volume.attachments,
                "status": volume.status,
            }
        return response

    def volume_show(self, name):
        """
        Show one volume
        """
        if self.volume_conn is None:
            raise SaltCloudSystemExit("No cinder endpoint available")

        volumes = self.volume_list(
            search_opts={"display_name": name},
        )
        volume = volumes[name]

        return volume

    def volume_create(
        self, name, size=100, snapshot=None, voltype=None, availability_zone=None
    ):
        """
        Create a block device
        """
        if self.volume_conn is None:
            raise SaltCloudSystemExit("No cinder endpoint available")
        nt_ks = self.volume_conn
        response = nt_ks.volumes.create(
            size=size,
            display_name=name,
            volume_type=voltype,
            snapshot_id=snapshot,
            availability_zone=availability_zone,
        )

        return self._volume_get(response.id)

    def volume_delete(self, name):
        """
        Delete a block device
        """
        if self.volume_conn is None:
            raise SaltCloudSystemExit("No cinder endpoint available")
        nt_ks = self.volume_conn
        try:
            volume = self.volume_show(name)
        except KeyError as exc:
            raise SaltCloudSystemExit("Unable to find {} volume: {}".format(name, exc))
        if volume["status"] == "deleted":
            return volume
        response = nt_ks.volumes.delete(volume["id"])
        return volume

    def volume_detach(self, name, timeout=300):
        """
        Detach a block device
        """
        try:
            volume = self.volume_show(name)
        except KeyError as exc:
            raise SaltCloudSystemExit("Unable to find {} volume: {}".format(name, exc))
        if not volume["attachments"]:
            return True
        response = self.compute_conn.volumes.delete_server_volume(
            volume["attachments"][0]["server_id"], volume["attachments"][0]["id"]
        )
        trycount = 0
        start = time.time()
        while True:
            trycount += 1
            try:
                response = self._volume_get(volume["id"])
                if response["status"] == "available":
                    return response
            except Exception as exc:  # pylint: disable=broad-except
                log.debug("Volume is detaching: %s", name)
                time.sleep(1)
                if time.time() - start > timeout:
                    log.error(
                        "Timed out after %d seconds while waiting for data", timeout
                    )
                    return False

                log.debug("Retrying volume_show() (try %d)", trycount)

    def volume_attach(self, name, server_name, device="/dev/xvdb", timeout=300):
        """
        Attach a block device
        """
        try:
            volume = self.volume_show(name)
        except KeyError as exc:
            raise SaltCloudSystemExit("Unable to find {} volume: {}".format(name, exc))
        server = self.server_by_name(server_name)
        response = self.compute_conn.volumes.create_server_volume(
            server.id, volume["id"], device=device
        )
        trycount = 0
        start = time.time()
        while True:
            trycount += 1
            try:
                response = self._volume_get(volume["id"])
                if response["status"] == "in-use":
                    return response
            except Exception as exc:  # pylint: disable=broad-except
                log.debug("Volume is attaching: %s", name)
                time.sleep(1)
                if time.time() - start > timeout:
                    log.error(
                        "Timed out after %s seconds while waiting for data", timeout
                    )
                    return False

                log.debug("Retrying volume_show() (try %s)", trycount)

    def suspend(self, instance_id):
        """
        Suspend a server
        """
        nt_ks = self.compute_conn
        response = nt_ks.servers.suspend(instance_id)
        return True

    def resume(self, instance_id):
        """
        Resume a server
        """
        nt_ks = self.compute_conn
        response = nt_ks.servers.resume(instance_id)
        return True

    def lock(self, instance_id):
        """
        Lock an instance
        """
        nt_ks = self.compute_conn
        response = nt_ks.servers.lock(instance_id)
        return True

    def delete(self, instance_id):
        """
        Delete a server
        """
        nt_ks = self.compute_conn
        response = nt_ks.servers.delete(instance_id)
        return True

    def flavor_list(self):
        """
        Return a list of available flavors (nova flavor-list)
        """
        nt_ks = self.compute_conn
        ret = {}
        for flavor in nt_ks.flavors.list():
            links = {}
            for link in flavor.links:
                links[link["rel"]] = link["href"]
            ret[flavor.name] = {
                "disk": flavor.disk,
                "id": flavor.id,
                "name": flavor.name,
                "ram": flavor.ram,
                "swap": flavor.swap,
                "vcpus": flavor.vcpus,
                "links": links,
            }
            if hasattr(flavor, "rxtx_factor"):
                ret[flavor.name]["rxtx_factor"] = flavor.rxtx_factor
        return ret

    list_sizes = flavor_list

    def flavor_create(
        self,
        name,  # pylint: disable=C0103
        flavor_id=0,  # pylint: disable=C0103
        ram=0,
        disk=0,
        vcpus=1,
    ):
        """
        Create a flavor
        """
        nt_ks = self.compute_conn
        nt_ks.flavors.create(
            name=name, flavorid=flavor_id, ram=ram, disk=disk, vcpus=vcpus
        )
        return {"name": name, "id": flavor_id, "ram": ram, "disk": disk, "vcpus": vcpus}

    def flavor_delete(self, flavor_id):  # pylint: disable=C0103
        """
        Delete a flavor
        """
        nt_ks = self.compute_conn
        nt_ks.flavors.delete(flavor_id)
        return "Flavor deleted: {}".format(flavor_id)

    def keypair_list(self):
        """
        List keypairs
        """
        nt_ks = self.compute_conn
        ret = {}
        for keypair in nt_ks.keypairs.list():
            ret[keypair.name] = {
                "name": keypair.name,
                "fingerprint": keypair.fingerprint,
                "public_key": keypair.public_key,
            }
        return ret

    def keypair_add(self, name, pubfile=None, pubkey=None):
        """
        Add a keypair
        """
        nt_ks = self.compute_conn
        if pubfile:
            with salt.utils.files.fopen(pubfile, "r") as fp_:
                pubkey = salt.utils.stringutils.to_unicode(fp_.read())
        if not pubkey:
            return False
        nt_ks.keypairs.create(name, public_key=pubkey)
        ret = {"name": name, "pubkey": pubkey}
        return ret

    def keypair_delete(self, name):
        """
        Delete a keypair
        """
        nt_ks = self.compute_conn
        nt_ks.keypairs.delete(name)
        return "Keypair deleted: {}".format(name)

    def image_show(self, image_id):
        """
        Show image details and metadata
        """
        nt_ks = self.compute_conn
        image = nt_ks.images.get(image_id)
        links = {}
        for link in image.links:
            links[link["rel"]] = link["href"]
        ret = {
            "name": image.name,
            "id": image.id,
            "status": image.status,
            "progress": image.progress,
            "created": image.created,
            "updated": image.updated,
            "metadata": image.metadata,
            "links": links,
        }
        if hasattr(image, "minDisk"):
            ret["minDisk"] = image.minDisk
        if hasattr(image, "minRam"):
            ret["minRam"] = image.minRam

        return ret

    def image_list(self, name=None):
        """
        List server images
        """
        nt_ks = self.compute_conn
        ret = {}
        for image in nt_ks.images.list():
            links = {}
            for link in image.links:
                links[link["rel"]] = link["href"]
            ret[image.name] = {
                "name": image.name,
                "id": image.id,
                "status": image.status,
                "progress": image.progress,
                "created": image.created,
                "updated": image.updated,
                "metadata": image.metadata,
                "links": links,
            }
            if hasattr(image, "minDisk"):
                ret[image.name]["minDisk"] = image.minDisk
            if hasattr(image, "minRam"):
                ret[image.name]["minRam"] = image.minRam
        if name:
            return {name: ret[name]}
        return ret

    list_images = image_list

    def image_meta_set(
        self, image_id=None, name=None, **kwargs
    ):  # pylint: disable=C0103
        """
        Set image metadata
        """
        nt_ks = self.compute_conn
        if name:
            for image in nt_ks.images.list():
                if image.name == name:
                    image_id = image.id  # pylint: disable=C0103
        if not image_id:
            return {"Error": "A valid image name or id was not specified"}
        nt_ks.images.set_meta(image_id, kwargs)
        return {image_id: kwargs}

    def image_meta_delete(
        self, image_id=None, name=None, keys=None  # pylint: disable=C0103
    ):
        """
        Delete image metadata
        """
        nt_ks = self.compute_conn
        if name:
            for image in nt_ks.images.list():
                if image.name == name:
                    image_id = image.id  # pylint: disable=C0103
        pairs = keys.split(",")
        if not image_id:
            return {"Error": "A valid image name or id was not specified"}
        nt_ks.images.delete_meta(image_id, pairs)
        return {image_id: "Deleted: {}".format(pairs)}

    def server_list(self):
        """
        List servers
        """
        nt_ks = self.compute_conn
        ret = {}
        for item in nt_ks.servers.list():
            try:
                ret[item.name] = {
                    "id": item.id,
                    "name": item.name,
                    "state": item.status,
                    "accessIPv4": item.accessIPv4,
                    "accessIPv6": item.accessIPv6,
                    "flavor": {"id": item.flavor["id"], "links": item.flavor["links"]},
                    "image": {
                        "id": item.image["id"] if item.image else "Boot From Volume",
                        "links": item.image["links"] if item.image else "",
                    },
                }
            except TypeError:
                pass
        return ret

    def server_list_min(self):
        """
        List minimal information about servers
        """
        nt_ks = self.compute_conn
        ret = {}
        for item in nt_ks.servers.list(detailed=False):
            try:
                ret[item.name] = {"id": item.id, "state": "Running"}
            except TypeError:
                pass
        return ret

    def server_list_detailed(self):
        """
        Detailed list of servers
        """
        nt_ks = self.compute_conn
        ret = {}
        for item in nt_ks.servers.list():
            try:
                ret[item.name] = {
                    "OS-EXT-SRV-ATTR": {},
                    "OS-EXT-STS": {},
                    "accessIPv4": item.accessIPv4,
                    "accessIPv6": item.accessIPv6,
                    "addresses": item.addresses,
                    "created": item.created,
                    "flavor": {"id": item.flavor["id"], "links": item.flavor["links"]},
                    "hostId": item.hostId,
                    "id": item.id,
                    "image": {
                        "id": item.image["id"] if item.image else "Boot From Volume",
                        "links": item.image["links"] if item.image else "",
                    },
                    "key_name": item.key_name,
                    "links": item.links,
                    "metadata": item.metadata,
                    "name": item.name,
                    "state": item.status,
                    "tenant_id": item.tenant_id,
                    "updated": item.updated,
                    "user_id": item.user_id,
                }
            except TypeError:
                continue

            ret[item.name]["progress"] = getattr(item, "progress", "0")

            if hasattr(item.__dict__, "OS-DCF:diskConfig"):
                ret[item.name]["OS-DCF"] = {
                    "diskConfig": item.__dict__["OS-DCF:diskConfig"]
                }
            if hasattr(item.__dict__, "OS-EXT-SRV-ATTR:host"):
                ret[item.name]["OS-EXT-SRV-ATTR"]["host"] = item.__dict__[
                    "OS-EXT-SRV-ATTR:host"
                ]
            if hasattr(item.__dict__, "OS-EXT-SRV-ATTR:hypervisor_hostname"):
                ret[item.name]["OS-EXT-SRV-ATTR"][
                    "hypervisor_hostname"
                ] = item.__dict__["OS-EXT-SRV-ATTR:hypervisor_hostname"]
            if hasattr(item.__dict__, "OS-EXT-SRV-ATTR:instance_name"):
                ret[item.name]["OS-EXT-SRV-ATTR"]["instance_name"] = item.__dict__[
                    "OS-EXT-SRV-ATTR:instance_name"
                ]
            if hasattr(item.__dict__, "OS-EXT-STS:power_state"):
                ret[item.name]["OS-EXT-STS"]["power_state"] = item.__dict__[
                    "OS-EXT-STS:power_state"
                ]
            if hasattr(item.__dict__, "OS-EXT-STS:task_state"):
                ret[item.name]["OS-EXT-STS"]["task_state"] = item.__dict__[
                    "OS-EXT-STS:task_state"
                ]
            if hasattr(item.__dict__, "OS-EXT-STS:vm_state"):
                ret[item.name]["OS-EXT-STS"]["vm_state"] = item.__dict__[
                    "OS-EXT-STS:vm_state"
                ]
            if hasattr(item.__dict__, "security_groups"):
                ret[item.name]["security_groups"] = item.__dict__["security_groups"]
        return ret

    def server_show(self, server_id):
        """
        Show details of one server
        """
        ret = {}
        try:
            servers = self.server_list_detailed()
        except AttributeError:
            raise SaltCloudSystemExit(
                "Corrupt server in server_list_detailed. Remove corrupt servers."
            )
        for server_name, server in servers.items():
            if str(server["id"]) == server_id:
                ret[server_name] = server
        return ret

    def secgroup_create(self, name, description):
        """
        Create a security group
        """
        nt_ks = self.compute_conn
        nt_ks.security_groups.create(name, description)
        ret = {"name": name, "description": description}
        return ret

    def secgroup_delete(self, name):
        """
        Delete a security group
        """
        nt_ks = self.compute_conn
        for item in nt_ks.security_groups.list():
            if item.name == name:
                nt_ks.security_groups.delete(item.id)
                return {name: "Deleted security group: {}".format(name)}
        return "Security group not found: {}".format(name)

    def secgroup_list(self):
        """
        List security groups
        """
        nt_ks = self.compute_conn
        ret = {}
        for item in nt_ks.security_groups.list():
            ret[item.name] = {
                "name": item.name,
                "description": item.description,
                "id": item.id,
                "tenant_id": item.tenant_id,
                "rules": item.rules,
            }
        return ret

    def _item_list(self):
        """
        List items
        """
        nt_ks = self.compute_conn
        ret = []
        for item in nt_ks.items.list():
            ret.append(item.__dict__)
        return ret

    def _network_show(self, name, network_lst):
        """
        Parse the returned network list
        """
        for net in network_lst:
            if net.label == name:
                return net.__dict__
        return {}

    def network_show(self, name):
        """
        Show network information
        """
        nt_ks = self.compute_conn
        net_list = nt_ks.networks.list()
        return self._network_show(name, net_list)

    def network_list(self):
        """
        List extra private networks
        """
        nt_ks = self.compute_conn
        return [network.__dict__ for network in nt_ks.networks.list()]

    def _sanatize_network_params(self, kwargs):
        """
        Sanatize novaclient network parameters
        """
        params = [
            "label",
            "bridge",
            "bridge_interface",
            "cidr",
            "cidr_v6",
            "dns1",
            "dns2",
            "fixed_cidr",
            "gateway",
            "gateway_v6",
            "multi_host",
            "priority",
            "project_id",
            "vlan_start",
            "vpn_start",
        ]

        for variable in list(kwargs):  # iterate over a copy, we might delete some
            if variable not in params:
                del kwargs[variable]
        return kwargs

    def network_create(self, name, **kwargs):
        """
        Create extra private network
        """
        nt_ks = self.compute_conn
        kwargs["label"] = name
        kwargs = self._sanatize_network_params(kwargs)
        net = nt_ks.networks.create(**kwargs)
        return net.__dict__

    def _server_uuid_from_name(self, name):
        """
        Get server uuid from name
        """
        return self.server_list().get(name, {}).get("id", "")

    def virtual_interface_list(self, name):
        """
        Get virtual interfaces on slice
        """
        nt_ks = self.compute_conn
        nets = nt_ks.virtual_interfaces.list(self._server_uuid_from_name(name))
        return [network.__dict__ for network in nets]

    def virtual_interface_create(self, name, net_name):
        """
        Add an interfaces to a slice
        """
        nt_ks = self.compute_conn
        serverid = self._server_uuid_from_name(name)
        networkid = self.network_show(net_name).get("id", None)
        if networkid is None:
            return {net_name: False}
        nets = nt_ks.virtual_interfaces.create(networkid, serverid)
        return nets

    def floating_ip_pool_list(self):
        """
        List all floating IP pools

        .. versionadded:: 2016.3.0
        """
        nt_ks = self.compute_conn
        pools = nt_ks.floating_ip_pools.list()
        response = {}
        for pool in pools:
            response[pool.name] = {
                "name": pool.name,
            }
        return response

    def floating_ip_list(self):
        """
        List floating IPs

        .. versionadded:: 2016.3.0
        """
        nt_ks = self.compute_conn
        floating_ips = nt_ks.floating_ips.list()
        response = {}
        for floating_ip in floating_ips:
            response[floating_ip.ip] = {
                "ip": floating_ip.ip,
                "fixed_ip": floating_ip.fixed_ip,
                "id": floating_ip.id,
                "instance_id": floating_ip.instance_id,
                "pool": floating_ip.pool,
            }
        return response

    def floating_ip_show(self, ip):
        """
        Show info on specific floating IP

        .. versionadded:: 2016.3.0
        """
        nt_ks = self.compute_conn
        floating_ips = nt_ks.floating_ips.list()
        for floating_ip in floating_ips:
            if floating_ip.ip == ip:
                response = {
                    "ip": floating_ip.ip,
                    "fixed_ip": floating_ip.fixed_ip,
                    "id": floating_ip.id,
                    "instance_id": floating_ip.instance_id,
                    "pool": floating_ip.pool,
                }
                return response
        return {}

    def floating_ip_create(self, pool=None):
        """
        Allocate a floating IP

        .. versionadded:: 2016.3.0
        """
        nt_ks = self.compute_conn
        floating_ip = nt_ks.floating_ips.create(pool)
        response = {
            "ip": floating_ip.ip,
            "fixed_ip": floating_ip.fixed_ip,
            "id": floating_ip.id,
            "instance_id": floating_ip.instance_id,
            "pool": floating_ip.pool,
        }
        return response

    def floating_ip_delete(self, floating_ip):
        """
        De-allocate a floating IP

        .. versionadded:: 2016.3.0
        """
        ip = self.floating_ip_show(floating_ip)
        nt_ks = self.compute_conn
        return nt_ks.floating_ips.delete(ip)

    def floating_ip_associate(self, server_name, floating_ip):
        """
        Associate floating IP address to server

        .. versionadded:: 2016.3.0
        """
        nt_ks = self.compute_conn
        server_ = self.server_by_name(server_name)
        server = nt_ks.servers.get(server_.__dict__["id"])
        server.add_floating_ip(floating_ip)
        return self.floating_ip_list()[floating_ip]

    def floating_ip_disassociate(self, server_name, floating_ip):
        """
        Disassociate a floating IP from server

        .. versionadded:: 2016.3.0
        """
        nt_ks = self.compute_conn
        server_ = self.server_by_name(server_name)
        server = nt_ks.servers.get(server_.__dict__["id"])
        server.remove_floating_ip(floating_ip)
        return self.floating_ip_list()[floating_ip]


# The following is a list of functions that need to be incorporated in the
# nova module. This list should be updated as functions are added.
#
# absolute-limits     Print a list of absolute limits for a user
# actions             Retrieve server actions.
# add-fixed-ip        Add new IP address to network.
# aggregate-add-host  Add the host to the specified aggregate.
# aggregate-create    Create a new aggregate with the specified details.
# aggregate-delete    Delete the aggregate by its id.
# aggregate-details   Show details of the specified aggregate.
# aggregate-list      Print a list of all aggregates.
# aggregate-remove-host
#                     Remove the specified host from the specified aggregate.
# aggregate-set-metadata
#                     Update the metadata associated with the aggregate.
# aggregate-update    Update the aggregate's name and optionally
#                     availability zone.
# cloudpipe-create    Create a cloudpipe instance for the given project
# cloudpipe-list      Print a list of all cloudpipe instances.
# console-log         Get console log output of a server.
# credentials         Show user credentials returned from auth
# describe-resource   Show details about a resource
# diagnostics         Retrieve server diagnostics.
# dns-create          Create a DNS entry for domain, name and ip.
# dns-create-private-domain
#                     Create the specified DNS domain.
# dns-create-public-domain
#                     Create the specified DNS domain.
# dns-delete          Delete the specified DNS entry.
# dns-delete-domain   Delete the specified DNS domain.
# dns-domains         Print a list of available dns domains.
# dns-list            List current DNS entries for domain and ip or domain
#                     and name.
# endpoints           Discover endpoints that get returned from the
#                     authenticate services
# get-vnc-console     Get a vnc console to a server.
# host-action         Perform a power action on a host.
# host-update         Update host settings.
# image-create        Create a new image by taking a snapshot of a running
#                     server.
# image-delete        Delete an image.
# live-migration      Migrates a running instance to a new machine.
# meta                Set or Delete metadata on a server.
# migrate             Migrate a server.
# pause               Pause a server.
# rate-limits         Print a list of rate limits for a user
# reboot              Reboot a server.
# rebuild             Shutdown, re-image, and re-boot a server.
# remove-fixed-ip     Remove an IP address from a server.
# rename              Rename a server.
# rescue              Rescue a server.
# resize              Resize a server.
# resize-confirm      Confirm a previous resize.
# resize-revert       Revert a previous resize (and return to the previous
#                     VM).
# root-password       Change the root password for a server.
# secgroup-add-group-rule
#                     Add a source group rule to a security group.
# secgroup-add-rule   Add a rule to a security group.
# secgroup-delete-group-rule
#                     Delete a source group rule from a security group.
# secgroup-delete-rule
#                     Delete a rule from a security group.
# secgroup-list-rules
#                     List rules for a security group.
# ssh                 SSH into a server.
# unlock              Unlock a server.
# unpause             Unpause a server.
# unrescue            Unrescue a server.
# usage-list          List usage data for all tenants
# volume-list         List all the volumes.
# volume-snapshot-create
#                     Add a new snapshot.
# volume-snapshot-delete
#                     Remove a snapshot.
# volume-snapshot-list
#                     List all the snapshots.
# volume-snapshot-show
#                     Show details about a snapshot.
# volume-type-create  Create a new volume type.
# volume-type-delete  Delete a specific flavor
# volume-type-list    Print a list of available 'volume types'.
# x509-create-cert    Create x509 cert for a user in tenant
# x509-get-root-cert  Fetches the x509 root cert.



"""
YAML Renderer for Salt

For YAML usage information see :ref:`Understanding YAML <yaml>`.
"""


import logging
import warnings

from yaml.constructor import ConstructorError
from yaml.parser import ParserError
from yaml.scanner import ScannerError

import salt.utils.url
import salt.utils.yamlloader as yamlloader_new
import salt.utils.yamlloader_old as yamlloader_old
from salt.exceptions import SaltRenderError
from salt.utils.odict import OrderedDict

log = logging.getLogger(__name__)

_ERROR_MAP = {
    "found character '\\t' that cannot start any token": "Illegal tab character"
}


def get_yaml_loader(argline):
    """
    Return the ordered dict yaml loader
    """

    def yaml_loader(*args):
        if __opts__.get("use_yamlloader_old"):
            yamlloader = yamlloader_old
        else:
            yamlloader = yamlloader_new
        return yamlloader.SaltYamlSafeLoader(*args, dictclass=OrderedDict)

    return yaml_loader


def render(yaml_data, saltenv="base", sls="", argline="", **kws):
    """
    Accepts YAML as a string or as a file object and runs it through the YAML
    parser.

    :rtype: A Python data structure
    """
    if __opts__.get("use_yamlloader_old"):
        log.warning(
            "Using the old YAML Loader for rendering, "
            "consider disabling this and using the tojson"
            " filter."
        )
        yamlloader = yamlloader_old
    else:
        yamlloader = yamlloader_new
    if not isinstance(yaml_data, str):
        yaml_data = yaml_data.read()
    with warnings.catch_warnings(record=True) as warn_list:
        try:
            data = yamlloader.load(yaml_data, Loader=get_yaml_loader(argline))
        except ScannerError as exc:
            err_type = _ERROR_MAP.get(exc.problem, exc.problem)
            line_num = exc.problem_mark.line + 1
            raise SaltRenderError(err_type, line_num, exc.problem_mark.buffer)
        except (ParserError, ConstructorError) as exc:
            raise SaltRenderError(exc)
        if len(warn_list) > 0:
            for item in warn_list:
                log.warning(
                    "%s found in %s saltenv=%s",
                    item.message,
                    salt.utils.url.create(sls),
                    saltenv,
                )
        if not data:
            data = {}

        def _validate_data(data):
            """
            PyYAML will for some reason allow improper YAML to be formed into
            an unhashable dict (that is, one with a dict as a key). This
            function will recursively go through and check the keys to make
            sure they're not dicts.
            """
            if isinstance(data, dict):
                for key, value in data.items():
                    if isinstance(key, dict):
                        raise SaltRenderError(
                            "Invalid YAML, possible double curly-brace"
                        )
                    _validate_data(value)
            elif isinstance(data, list):
                for item in data:
                    _validate_data(item)

        _validate_data(data)
        return data



"""
Configuration of the kernel using sysfs
=======================================

Control the kernel object attributes exported by sysfs

.. code-block:: yaml

  kernel/mm/transparent_hugepage/enabled
    sysfs.present:
      - value: never

.. versionadded:: 3006.0
"""

import re


def __virtual__():
    """
    This state is only available on Minions which support sysctl
    """
    if "sysfs.attr" in __salt__:
        return True
    return (False, "sysfs module could not be loaded")


def present(name, value, config=None):
    """
    Ensure that the named sysfs attribute is set with the defined value

    name
        The name of the sysfs attribute to edit

    value
        The sysfs value to apply

    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    current = __salt__["sysfs.read"](name)
    if current is False:
        ret["result"] = False
        ret["comment"] = "SysFS attribute {} doesn't exist.".format(name)
    else:
        # if the return is a dict, the "name" is an object not an attribute
        if isinstance(current, dict):
            ret["result"] = False
            ret["comment"] = "{} is not a SysFS attribute.".format(name)
        else:
            # some attribute files lists all available options and the selected one between []
            if isinstance(current, str):
                current = re.sub(r"(.*\[|\].*)", "", current)
            if value == current:
                ret["result"] = True
                ret["comment"] = "SysFS attribute {} is already set.".format(name)
            else:
                ret["result"] = None

    if ret["result"] is None:
        if __opts__["test"]:
            ret["comment"] = "SysFS attribute {} set to be changed.".format(name)
        else:
            update = __salt__["sysfs.write"](name, value)
            if not update:
                ret["result"] = False
                ret["comment"] = "Failed to set {} to {}".format(name, value)
            else:
                ret["result"] = True
                ret["changes"] = {name: value}
                ret["comment"] = "Updated SysFS attribute {} to {}".format(name, value)

    return ret



"""
Add all extra minion data to the pillar.

:codeauthor: Alexandru.Bleotu@morganstanley.ms.com

One can filter on the keys to include in the pillar by using the ``include``
parameter. For subkeys the ':' notation is supported (i.e. 'key:subkey')
The keyword ``<all>`` includes all keys.

Complete example in etc/salt/master
=====================================

.. code-block:: none

    ext_pillar:
      - extra_minion_data_in_pillar:
          include: *

    ext_pillar:
      - extra_minion_data_in_pillar:
          include:
              - key1
              - key2:subkey2

    ext_pillar:
      - extra_minion_data_in_pillar:
          include: <all>

"""
import logging

# Set up logging
log = logging.getLogger(__name__)

__virtualname__ = "extra_minion_data_in_pillar"


def __virtual__():
    return __virtualname__


def ext_pillar(minion_id, pillar, include, extra_minion_data=None):
    def get_subtree(key, source_dict):
        """
        Returns a subtree corresponfing to the specified key.

        key
            Key. Supports the ':' notation (e.g. 'key:subkey')

        source_dict
            Source dictionary
        """
        ret_dict = aux_dict = {}
        subtree = source_dict
        subkeys = key.split(":")
        # Build an empty intermediate subtree following the subkeys
        for subkey in subkeys[:-1]:
            # The result will be built in aux_dict
            aux_dict[subkey] = {}
            aux_dict = aux_dict[subkey]
            if subkey not in subtree:
                # The subkey is not in
                return {}
            subtree = subtree[subkey]
        if subkeys[-1] not in subtree:
            # Final subkey is not in subtree
            return {}
        # Assign the subtree value to the result
        aux_dict[subkeys[-1]] = subtree[subkeys[-1]]
        return ret_dict

    log.trace("minion_id = %s", minion_id)
    log.trace("include = %s", include)
    log.trace("extra_minion_data = %s", extra_minion_data)
    data = {}

    if not extra_minion_data:
        return {}
    if include in ["*", "<all>"]:
        return extra_minion_data
    data = {}
    for key in include:
        data.update(get_subtree(key, extra_minion_data))
    return data



"""
The networking module for SUSE based distros

.. versionadded:: 3005

"""

import logging
import os

import jinja2
import jinja2.exceptions

import salt.utils.files
import salt.utils.stringutils
import salt.utils.templates
import salt.utils.validate.net
from salt.exceptions import CommandExecutionError

# Set up logging
log = logging.getLogger(__name__)

# Set up template environment
JINJA = jinja2.Environment(
    loader=jinja2.FileSystemLoader(
        os.path.join(salt.utils.templates.TEMPLATE_DIRNAME, "suse_ip")
    )
)

# Define the module's virtual name
__virtualname__ = "ip"

# Default values for bonding
_BOND_DEFAULTS = {
    # 803.ad aggregation selection logic
    # 0 for stable (default)
    # 1 for bandwidth
    # 2 for count
    "ad_select": "0",
    # Max number of transmit queues (default = 16)
    "tx_queues": "16",
    # lacp_rate 0: Slow - every 30 seconds
    # lacp_rate 1: Fast - every 1 second
    "lacp_rate": "0",
    # Max bonds for this driver
    "max_bonds": "1",
    # Used with miimon.
    # On: driver sends mii
    # Off: ethtool sends mii
    "use_carrier": "0",
    # Default. Don't change unless you know what you are doing.
    "xmit_hash_policy": "layer2",
}
_SUSE_NETWORK_SCRIPT_DIR = "/etc/sysconfig/network"
_SUSE_NETWORK_FILE = "/etc/sysconfig/network/config"
_SUSE_NETWORK_ROUTES_FILE = "/etc/sysconfig/network/routes"
_CONFIG_TRUE = ("yes", "on", "true", "1", True)
_CONFIG_FALSE = ("no", "off", "false", "0", False)
_IFACE_TYPES = (
    "eth",
    "bond",
    "alias",
    "clone",
    "ipsec",
    "dialup",
    "bridge",
    "slave",
    "vlan",
    "ipip",
    "ib",
)


def __virtual__():
    """
    Confine this module to SUSE based distros
    """
    if __grains__["os_family"] == "Suse":
        return __virtualname__
    return (
        False,
        "The suse_ip execution module cannot be loaded: "
        "this module is only available on SUSE based distributions.",
    )


def _error_msg_iface(iface, option, expected):
    """
    Build an appropriate error message from a given option and
    a list of expected values.
    """
    if isinstance(expected, str):
        expected = (expected,)
    msg = "Invalid option -- Interface: {}, Option: {}, Expected: [{}]"
    return msg.format(iface, option, "|".join(str(e) for e in expected))


def _error_msg_routes(iface, option, expected):
    """
    Build an appropriate error message from a given option and
    a list of expected values.
    """
    msg = "Invalid option -- Route interface: {}, Option: {}, Expected: [{}]"
    return msg.format(iface, option, expected)


def _log_default_iface(iface, opt, value):
    log.info(
        "Using default option -- Interface: %s Option: %s Value: %s", iface, opt, value
    )


def _error_msg_network(option, expected):
    """
    Build an appropriate error message from a given option and
    a list of expected values.
    """
    if isinstance(expected, str):
        expected = (expected,)
    msg = "Invalid network setting -- Setting: {}, Expected: [{}]"
    return msg.format(option, "|".join(str(e) for e in expected))


def _log_default_network(opt, value):
    log.info("Using existing setting -- Setting: %s Value: %s", opt, value)


def _parse_suse_config(path):
    suse_config = _read_file(path)
    cv_suse_config = {}
    if suse_config:
        for line in suse_config:
            line = line.strip()
            if not line or line.startswith("!") or line.startswith("#"):
                continue
            pair = [p.rstrip() for p in line.split("=", 1)]
            if len(pair) != 2:
                continue
            name, value = pair
            cv_suse_config[name.upper()] = salt.utils.stringutils.dequote(value)

    return cv_suse_config


def _parse_ethtool_opts(opts, iface):
    """
    Parses valid options for ETHTOOL_OPTIONS of the interface
    Logs the error and raises AttributeError in case of getting invalid options
    """
    config = {}

    if "autoneg" in opts:
        if opts["autoneg"] in _CONFIG_TRUE:
            config.update({"autoneg": "on"})
        elif opts["autoneg"] in _CONFIG_FALSE:
            config.update({"autoneg": "off"})
        else:
            _raise_error_iface(iface, "autoneg", _CONFIG_TRUE + _CONFIG_FALSE)

    if "duplex" in opts:
        valid = ["full", "half"]
        if opts["duplex"] in valid:
            config.update({"duplex": opts["duplex"]})
        else:
            _raise_error_iface(iface, "duplex", valid)

    if "speed" in opts:
        valid = ["10", "100", "1000", "10000"]
        if str(opts["speed"]) in valid:
            config.update({"speed": opts["speed"]})
        else:
            _raise_error_iface(iface, opts["speed"], valid)

    if "advertise" in opts:
        valid = [
            "0x001",
            "0x002",
            "0x004",
            "0x008",
            "0x010",
            "0x020",
            "0x20000",
            "0x8000",
            "0x1000",
            "0x40000",
            "0x80000",
            "0x200000",
            "0x400000",
            "0x800000",
            "0x1000000",
            "0x2000000",
            "0x4000000",
        ]
        if str(opts["advertise"]) in valid:
            config.update({"advertise": opts["advertise"]})
        else:
            _raise_error_iface(iface, "advertise", valid)

    if "channels" in opts:
        channels_cmd = "-L {}".format(iface.strip())
        channels_params = []
        for option in ("rx", "tx", "other", "combined"):
            if option in opts["channels"]:
                valid = range(1, __grains__["num_cpus"] + 1)
                if opts["channels"][option] in valid:
                    channels_params.append(
                        "{} {}".format(option, opts["channels"][option])
                    )
                else:
                    _raise_error_iface(iface, opts["channels"][option], valid)
        if channels_params:
            config.update({channels_cmd: " ".join(channels_params)})

    valid = _CONFIG_TRUE + _CONFIG_FALSE
    for option in ("rx", "tx", "sg", "tso", "ufo", "gso", "gro", "lro"):
        if option in opts:
            if opts[option] in _CONFIG_TRUE:
                config.update({option: "on"})
            elif opts[option] in _CONFIG_FALSE:
                config.update({option: "off"})
            else:
                _raise_error_iface(iface, option, valid)

    return config


def _parse_settings_bond(opts, iface):
    """
    Parses valid options for bonding interface
    Logs the error and raises AttributeError in case of getting invalid options
    """
    if opts["mode"] in ("balance-rr", "0"):
        log.info("Device: %s Bonding Mode: load balancing (round-robin)", iface)
        return _parse_settings_bond_0(opts, iface)
    elif opts["mode"] in ("active-backup", "1"):
        log.info("Device: %s Bonding Mode: fault-tolerance (active-backup)", iface)
        return _parse_settings_bond_1(opts, iface)
    elif opts["mode"] in ("balance-xor", "2"):
        log.info("Device: %s Bonding Mode: load balancing (xor)", iface)
        return _parse_settings_bond_2(opts, iface)
    elif opts["mode"] in ("broadcast", "3"):
        log.info("Device: %s Bonding Mode: fault-tolerance (broadcast)", iface)
        return _parse_settings_bond_3(opts, iface)
    elif opts["mode"] in ("802.3ad", "4"):
        log.info(
            "Device: %s Bonding Mode: IEEE 802.3ad Dynamic link aggregation", iface
        )
        return _parse_settings_bond_4(opts, iface)
    elif opts["mode"] in ("balance-tlb", "5"):
        log.info("Device: %s Bonding Mode: transmit load balancing", iface)
        return _parse_settings_bond_5(opts, iface)
    elif opts["mode"] in ("balance-alb", "6"):
        log.info("Device: %s Bonding Mode: adaptive load balancing", iface)
        return _parse_settings_bond_6(opts, iface)
    else:
        valid = (
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "balance-rr",
            "active-backup",
            "balance-xor",
            "broadcast",
            "802.3ad",
            "balance-tlb",
            "balance-alb",
        )
        _raise_error_iface(iface, "mode", valid)


def _parse_settings_miimon(opts, iface):
    """
    Add shared settings for miimon support used by balance-rr, balance-xor
    bonding types.
    """
    ret = {}
    for binding in ("miimon", "downdelay", "updelay"):
        if binding in opts:
            try:
                int(opts[binding])
                ret.update({binding: opts[binding]})
            except ValueError:
                _raise_error_iface(iface, binding, "integer")

    if "miimon" in opts and "downdelay" not in opts:
        ret["downdelay"] = ret["miimon"] * 2

    if "miimon" in opts:
        if not opts["miimon"]:
            _raise_error_iface(iface, "miimon", "nonzero integer")

        for binding in ("downdelay", "updelay"):
            if binding in ret:
                if ret[binding] % ret["miimon"]:
                    _raise_error_iface(
                        iface,
                        binding,
                        "0 or a multiple of miimon ({})".format(ret["miimon"]),
                    )

        if "use_carrier" in opts:
            if opts["use_carrier"] in _CONFIG_TRUE:
                ret.update({"use_carrier": "1"})
            elif opts["use_carrier"] in _CONFIG_FALSE:
                ret.update({"use_carrier": "0"})
            else:
                valid = _CONFIG_TRUE + _CONFIG_FALSE
                _raise_error_iface(iface, "use_carrier", valid)
        else:
            _log_default_iface(iface, "use_carrier", _BOND_DEFAULTS["use_carrier"])
            ret.update({"use_carrier": _BOND_DEFAULTS["use_carrier"]})

    return ret


def _parse_settings_arp(opts, iface):
    """
    Add shared settings for arp used by balance-rr, balance-xor bonding types.
    """
    ret = {}
    if "arp_interval" in opts:
        try:
            int(opts["arp_interval"])
            ret.update({"arp_interval": opts["arp_interval"]})
        except ValueError:
            _raise_error_iface(iface, "arp_interval", "integer")

        # ARP targets in n.n.n.n form
        valid = "list of ips (up to 16)"
        if "arp_ip_target" in opts:
            if isinstance(opts["arp_ip_target"], list):
                if 1 <= len(opts["arp_ip_target"]) <= 16:
                    ret.update({"arp_ip_target": ",".join(opts["arp_ip_target"])})
                else:
                    _raise_error_iface(iface, "arp_ip_target", valid)
            else:
                _raise_error_iface(iface, "arp_ip_target", valid)
        else:
            _raise_error_iface(iface, "arp_ip_target", valid)

    return ret


def _parse_settings_bond_0(opts, iface):
    """
    Parses valid options for balance-rr (type 0) bonding interface
    Logs the error and raises AttributeError in case of getting invalid options
    """
    bond = {"mode": "0"}
    bond.update(_parse_settings_miimon(opts, iface))
    bond.update(_parse_settings_arp(opts, iface))

    if "miimon" not in opts and "arp_interval" not in opts:
        _raise_error_iface(
            iface, "miimon or arp_interval", "at least one of these is required"
        )

    return bond


def _parse_settings_bond_1(opts, iface):
    """
    Parses valid options for active-backup (type 1) bonding interface
    Logs the error and raises AttributeError in case of getting invalid options
    """
    bond = {"mode": "1"}
    bond.update(_parse_settings_miimon(opts, iface))

    if "miimon" not in opts:
        _raise_error_iface(iface, "miimon", "integer")

    if "primary" in opts:
        bond.update({"primary": opts["primary"]})

    return bond


def _parse_settings_bond_2(opts, iface):
    """
    Parses valid options for balance-xor (type 2) bonding interface
    Logs the error and raises AttributeError in case of getting invalid options
    """
    bond = {"mode": "2"}
    bond.update(_parse_settings_miimon(opts, iface))
    bond.update(_parse_settings_arp(opts, iface))

    if "miimon" not in opts and "arp_interval" not in opts:
        _raise_error_iface(
            iface, "miimon or arp_interval", "at least one of these is required"
        )

    if "hashing-algorithm" in opts:
        valid = ("layer2", "layer2+3", "layer3+4")
        if opts["hashing-algorithm"] in valid:
            bond.update({"xmit_hash_policy": opts["hashing-algorithm"]})
        else:
            _raise_error_iface(iface, "hashing-algorithm", valid)

    return bond


def _parse_settings_bond_3(opts, iface):
    """
    Parses valid options for broadcast (type 3) bonding interface
    Logs the error and raises AttributeError in case of getting invalid options
    """
    bond = {"mode": "3"}
    bond.update(_parse_settings_miimon(opts, iface))

    if "miimon" not in opts:
        _raise_error_iface(iface, "miimon", "integer")

    return bond


def _parse_settings_bond_4(opts, iface):
    """
    Parses valid options for 802.3ad (type 4) bonding interface
    Logs the error and raises AttributeError in case of getting invalid options
    """
    bond = {"mode": "4"}
    bond.update(_parse_settings_miimon(opts, iface))

    if "miimon" not in opts:
        _raise_error_iface(iface, "miimon", "integer")

    for binding in ("lacp_rate", "ad_select"):
        if binding in opts:
            if binding == "lacp_rate":
                valid = ("fast", "1", "slow", "0")
                if opts[binding] not in valid:
                    _raise_error_iface(iface, binding, valid)
                if opts[binding] == "fast":
                    opts.update({binding: "1"})
                if opts[binding] == "slow":
                    opts.update({binding: "0"})
            else:
                valid = "integer"
            try:
                int(opts[binding])
                bond.update({binding: opts[binding]})
            except ValueError:
                _raise_error_iface(iface, binding, valid)
        else:
            _log_default_iface(iface, binding, _BOND_DEFAULTS[binding])
            bond.update({binding: _BOND_DEFAULTS[binding]})

    if "hashing-algorithm" in opts:
        valid = ("layer2", "layer2+3", "layer3+4")
        if opts["hashing-algorithm"] in valid:
            bond.update({"xmit_hash_policy": opts["hashing-algorithm"]})
        else:
            _raise_error_iface(iface, "hashing-algorithm", valid)

    return bond


def _parse_settings_bond_5(opts, iface):
    """
    Parses valid options for balance-tlb (type 5) bonding interface
    Logs the error and raises AttributeError in case of getting invalid options
    """
    bond = {"mode": "5"}
    bond.update(_parse_settings_miimon(opts, iface))

    if "miimon" not in opts:
        _raise_error_iface(iface, "miimon", "integer")

    if "primary" in opts:
        bond.update({"primary": opts["primary"]})

    return bond


def _parse_settings_bond_6(opts, iface):
    """
    Parses valid options for balance-alb (type 6) bonding interface
    Logs the error and raises AttributeError in case of getting invalid options
    """
    bond = {"mode": "6"}
    bond.update(_parse_settings_miimon(opts, iface))

    if "miimon" not in opts:
        _raise_error_iface(iface, "miimon", "integer")

    if "primary" in opts:
        bond.update({"primary": opts["primary"]})

    return bond


def _parse_settings_vlan(opts, iface):
    """
    Filters given options and outputs valid settings for a vlan
    """
    vlan = {}
    if "reorder_hdr" in opts:
        if opts["reorder_hdr"] in _CONFIG_TRUE + _CONFIG_FALSE:
            vlan.update({"reorder_hdr": opts["reorder_hdr"]})
        else:
            valid = _CONFIG_TRUE + _CONFIG_FALSE
            _raise_error_iface(iface, "reorder_hdr", valid)

    if "vlan_id" in opts:
        if opts["vlan_id"] > 0:
            vlan.update({"vlan_id": opts["vlan_id"]})
        else:
            _raise_error_iface(iface, "vlan_id", "Positive integer")

    if "phys_dev" in opts:
        if len(opts["phys_dev"]) > 0:
            vlan.update({"phys_dev": opts["phys_dev"]})
        else:
            _raise_error_iface(iface, "phys_dev", "Non-empty string")

    return vlan


def _parse_settings_eth(opts, iface_type, enabled, iface):
    """
    Filters given options and outputs valid settings for a
    network interface.
    """
    result = {"name": iface}
    if "proto" in opts:
        valid = [
            "static",
            "dhcp",
            "dhcp4",
            "dhcp6",
            "autoip",
            "dhcp+autoip",
            "auto6",
            "6to4",
            "none",
        ]
        if opts["proto"] in valid:
            result["proto"] = opts["proto"]
        else:
            _raise_error_iface(iface, opts["proto"], valid)

    if "mtu" in opts:
        try:
            result["mtu"] = int(opts["mtu"])
        except ValueError:
            _raise_error_iface(iface, "mtu", ["integer"])

    if "hwaddr" in opts and "macaddr" in opts:
        msg = "Cannot pass both hwaddr and macaddr. Must use either hwaddr or macaddr"
        log.error(msg)
        raise AttributeError(msg)

    if iface_type not in ("bridge",):
        ethtool = _parse_ethtool_opts(opts, iface)
        if ethtool:
            result["ethtool"] = " ".join(
                ["{} {}".format(x, y) for x, y in ethtool.items()]
            )

    if iface_type == "slave":
        result["proto"] = "none"

    if iface_type == "bond":
        if "mode" not in opts:
            msg = "Missing required option 'mode'"
            log.error("%s for bond interface '%s'", msg, iface)
            raise AttributeError(msg)
        bonding = _parse_settings_bond(opts, iface)
        if bonding:
            result["bonding"] = " ".join(
                ["{}={}".format(x, y) for x, y in bonding.items()]
            )
            result["devtype"] = "Bond"
            if "slaves" in opts:
                if isinstance(opts["slaves"], list):
                    result["slaves"] = opts["slaves"]
                else:
                    result["slaves"] = opts["slaves"].split()

    if iface_type == "vlan":
        vlan = _parse_settings_vlan(opts, iface)
        if vlan:
            result["devtype"] = "Vlan"
            for opt in vlan:
                result[opt] = opts[opt]

    if iface_type == "eth":
        result["devtype"] = "Ethernet"

    if iface_type == "bridge":
        result["devtype"] = "Bridge"
        bypassfirewall = True
        valid = _CONFIG_TRUE + _CONFIG_FALSE
        for opt in ("bypassfirewall",):
            if opt in opts:
                if opts[opt] in _CONFIG_TRUE:
                    bypassfirewall = True
                elif opts[opt] in _CONFIG_FALSE:
                    bypassfirewall = False
                else:
                    _raise_error_iface(iface, opts[opt], valid)

        bridgectls = [
            "net.bridge.bridge-nf-call-ip6tables",
            "net.bridge.bridge-nf-call-iptables",
            "net.bridge.bridge-nf-call-arptables",
        ]

        if bypassfirewall:
            sysctl_value = 0
        else:
            sysctl_value = 1

        for sysctl in bridgectls:
            try:
                __salt__["sysctl.persist"](sysctl, sysctl_value)
            except CommandExecutionError:
                log.warning("Failed to set sysctl: %s", sysctl)

    else:
        if "bridge" in opts:
            result["bridge"] = opts["bridge"]

    if iface_type == "ipip":
        result["devtype"] = "IPIP"
        for opt in ("my_inner_ipaddr", "my_outer_ipaddr"):
            if opt not in opts:
                _raise_error_iface(iface, opt, "1.2.3.4")
            else:
                result[opt] = opts[opt]
    if iface_type == "ib":
        result["devtype"] = "InfiniBand"

    if "prefix" in opts:
        if "netmask" in opts:
            msg = "Cannot use prefix and netmask together"
            log.error(msg)
            raise AttributeError(msg)
        result["prefix"] = opts["prefix"]
    elif "netmask" in opts:
        result["netmask"] = opts["netmask"]

    for opt in (
        "ipaddr",
        "master",
        "srcaddr",
        "delay",
        "domain",
        "gateway",
        "uuid",
        "nickname",
        "zone",
    ):
        if opt in opts:
            result[opt] = opts[opt]

    if "ipaddrs" in opts or "ipv6addr" in opts or "ipv6addrs" in opts:
        result["ipaddrs"] = []
    if "ipaddrs" in opts:
        for opt in opts["ipaddrs"]:
            if salt.utils.validate.net.ipv4_addr(
                opt
            ) or salt.utils.validate.net.ipv6_addr(opt):
                result["ipaddrs"].append(opt)
            else:
                msg = "{} is invalid ipv4 or ipv6 CIDR".format(opt)
                log.error(msg)
                raise AttributeError(msg)
    if "ipv6addr" in opts:
        if salt.utils.validate.net.ipv6_addr(opts["ipv6addr"]):
            result["ipaddrs"].append(opts["ipv6addr"])
        else:
            msg = "{} is invalid ipv6 CIDR".format(opt)
            log.error(msg)
            raise AttributeError(msg)
    if "ipv6addrs" in opts:
        for opt in opts["ipv6addrs"]:
            if salt.utils.validate.net.ipv6_addr(opt):
                result["ipaddrs"].append(opt)
            else:
                msg = "{} is invalid ipv6 CIDR".format(opt)
                log.error(msg)
                raise AttributeError(msg)

    if "enable_ipv6" in opts:
        result["enable_ipv6"] = opts["enable_ipv6"]

    valid = _CONFIG_TRUE + _CONFIG_FALSE
    for opt in (
        "onparent",
        "peerdns",
        "peerroutes",
        "slave",
        "vlan",
        "defroute",
        "stp",
        "ipv6_peerdns",
        "ipv6_defroute",
        "ipv6_peerroutes",
        "ipv6_autoconf",
        "ipv4_failure_fatal",
        "dhcpv6c",
    ):
        if opt in opts:
            if opts[opt] in _CONFIG_TRUE:
                result[opt] = "yes"
            elif opts[opt] in _CONFIG_FALSE:
                result[opt] = "no"
            else:
                _raise_error_iface(iface, opts[opt], valid)

    if "onboot" in opts:
        log.warning(
            "The 'onboot' option is controlled by the 'enabled' option. "
            "Interface: %s Enabled: %s",
            iface,
            enabled,
        )

    if "startmode" in opts:
        valid = ("manual", "auto", "nfsroot", "hotplug", "off")
        if opts["startmode"] in valid:
            result["startmode"] = opts["startmode"]
        else:
            _raise_error_iface(iface, opts["startmode"], valid)
    else:
        if enabled:
            result["startmode"] = "auto"
        else:
            result["startmode"] = "off"

    # This vlan is in opts, and should be only used in range interface
    # will affect jinja template for interface generating
    if "vlan" in opts:
        if opts["vlan"] in _CONFIG_TRUE:
            result["vlan"] = "yes"
        elif opts["vlan"] in _CONFIG_FALSE:
            result["vlan"] = "no"
        else:
            _raise_error_iface(iface, opts["vlan"], valid)

    if "arpcheck" in opts:
        if opts["arpcheck"] in _CONFIG_FALSE:
            result["arpcheck"] = "no"

    if "ipaddr_start" in opts:
        result["ipaddr_start"] = opts["ipaddr_start"]

    if "ipaddr_end" in opts:
        result["ipaddr_end"] = opts["ipaddr_end"]

    if "clonenum_start" in opts:
        result["clonenum_start"] = opts["clonenum_start"]

    if "hwaddr" in opts:
        result["hwaddr"] = opts["hwaddr"]

    if "macaddr" in opts:
        result["macaddr"] = opts["macaddr"]

    # If NetworkManager is available, we can control whether we use
    # it or not
    if "nm_controlled" in opts:
        if opts["nm_controlled"] in _CONFIG_TRUE:
            result["nm_controlled"] = "yes"
        elif opts["nm_controlled"] in _CONFIG_FALSE:
            result["nm_controlled"] = "no"
        else:
            _raise_error_iface(iface, opts["nm_controlled"], valid)
    else:
        result["nm_controlled"] = "no"

    return result


def _parse_routes(iface, opts):
    """
    Filters given options and outputs valid settings for
    the route settings file.
    """
    # Normalize keys
    opts = {k.lower(): v for (k, v) in opts.items()}
    result = {}
    if "routes" not in opts:
        _raise_error_routes(iface, "routes", "List of routes")

    for opt in opts:
        result[opt] = opts[opt]

    return result


def _parse_network_settings(opts, current):
    """
    Filters given options and outputs valid settings for
    the global network settings file.
    """
    # Normalize keys
    opts = {k.lower(): v for (k, v) in opts.items()}
    current = {k.lower(): v for (k, v) in current.items()}

    # Check for supported parameters
    retain_settings = opts.get("retain_settings", False)
    result = {}
    if retain_settings:
        for opt in current:
            nopt = opt
            if opt == "netconfig_dns_static_servers":
                nopt = "dns"
                result[nopt] = current[opt].split()
            elif opt == "netconfig_dns_static_searchlist":
                nopt = "dns_search"
                result[nopt] = current[opt].split()
            elif opt.startswith("netconfig_") and opt not in (
                "netconfig_modules_order",
                "netconfig_verbose",
                "netconfig_force_replace",
            ):
                nopt = opt[10:]
                result[nopt] = current[opt]
            else:
                result[nopt] = current[opt]
            _log_default_network(nopt, current[opt])

    for opt in opts:
        if opt in ("dns", "dns_search") and not isinstance(opts[opt], list):
            result[opt] = opts[opt].split()
        else:
            result[opt] = opts[opt]
    return result


def _raise_error_iface(iface, option, expected):
    """
    Log and raise an error with a logical formatted message.
    """
    msg = _error_msg_iface(iface, option, expected)
    log.error(msg)
    raise AttributeError(msg)


def _raise_error_network(option, expected):
    """
    Log and raise an error with a logical formatted message.
    """
    msg = _error_msg_network(option, expected)
    log.error(msg)
    raise AttributeError(msg)


def _raise_error_routes(iface, option, expected):
    """
    Log and raise an error with a logical formatted message.
    """
    msg = _error_msg_routes(iface, option, expected)
    log.error(msg)
    raise AttributeError(msg)


def _read_file(path):
    """
    Reads and returns the contents of a file
    """
    try:
        with salt.utils.files.fopen(path, "rb") as rfh:
            return _get_non_blank_lines(salt.utils.stringutils.to_unicode(rfh.read()))
    except OSError:
        return []  # Return empty list for type consistency


def _write_file_iface(iface, data, folder, pattern):
    """
    Writes a file to disk
    """
    filename = os.path.join(folder, pattern.format(iface))
    if not os.path.exists(folder):
        msg = "{} cannot be written. {} does not exist".format(filename, folder)
        log.error(msg)
        raise AttributeError(msg)
    with salt.utils.files.fopen(filename, "w") as fp_:
        fp_.write(salt.utils.stringutils.to_str(data))


def _write_file_network(data, filename):
    """
    Writes a file to disk
    """
    with salt.utils.files.fopen(filename, "w") as fp_:
        fp_.write(salt.utils.stringutils.to_str(data))


def _get_non_blank_lines(data):
    lines = data.splitlines()
    try:  # Discard newlines if they exist
        lines.remove("")
    except ValueError:
        pass
    return lines


def build_interface(iface, iface_type, enabled, **settings):
    """
    Build an interface script for a network interface.

    Args:
        :param iface:
            The name of the interface to build the configuration for

        :param iface_type:
            The type of the interface. The following types are possible:
              - eth
              - bond
              - alias
              - clone
              - ipsec
              - dialup
              - bridge
              - slave
              - vlan
              - ipip
              - ib

        :param enabled:
            Build the interface enabled or disabled

        :param settings:
            The settings for the interface

    Returns:
        dict: A dictionary of file/content

    CLI Example:

    .. code-block:: bash

        salt '*' ip.build_interface eth0 eth <settings>
    """
    iface_type = iface_type.lower()

    if iface_type not in _IFACE_TYPES:
        _raise_error_iface(iface, iface_type, _IFACE_TYPES)

    if iface_type == "slave":
        settings["slave"] = "yes"
        if "master" not in settings:
            msg = "master is a required setting for slave interfaces"
            log.error(msg)
            raise AttributeError(msg)

    if iface_type == "bond":
        if "mode" not in settings:
            msg = "mode is required for bond interfaces"
            log.error(msg)
            raise AttributeError(msg)
        settings["mode"] = str(settings["mode"])

    if iface_type == "vlan":
        settings["vlan"] = "yes"

    if iface_type == "bridge" and not __salt__["pkg.version"]("bridge-utils"):
        __salt__["pkg.install"]("bridge-utils")

    if iface_type in (
        "eth",
        "bond",
        "bridge",
        "slave",
        "vlan",
        "ipip",
        "ib",
        "alias",
    ):
        opts = _parse_settings_eth(settings, iface_type, enabled, iface)
        try:
            template = JINJA.get_template("ifcfg.jinja")
        except jinja2.exceptions.TemplateNotFound:
            log.error("Could not load template ifcfg.jinja")
            return ""
        log.debug("Interface opts:\n%s", opts)
        ifcfg = template.render(opts)

    if settings.get("test"):
        return _get_non_blank_lines(ifcfg)

    _write_file_iface(iface, ifcfg, _SUSE_NETWORK_SCRIPT_DIR, "ifcfg-{}")
    path = os.path.join(_SUSE_NETWORK_SCRIPT_DIR, "ifcfg-{}".format(iface))

    return _read_file(path)


def build_routes(iface, **settings):
    """
    Build a route script for a network interface.

    Args:
        :param iface:
            The name of the interface to build the routes for

        :param settings:
            The settings for the routes

    Returns:
        dict: A dictionary of file/content

    CLI Example:

    .. code-block:: bash

        salt '*' ip.build_routes eth0 <settings>
    """

    template = "ifroute.jinja"
    log.debug("Template name: %s", template)

    opts = _parse_routes(iface, settings)
    log.debug("Opts:\n%s", opts)
    try:
        template = JINJA.get_template(template)
    except jinja2.exceptions.TemplateNotFound:
        log.error("Could not load template %s", template)
        return ""
    log.debug("IP routes:\n%s", opts["routes"])

    if iface == "routes":
        routecfg = template.render(routes=opts["routes"])
    else:
        routecfg = template.render(routes=opts["routes"], iface=iface)

    if settings["test"]:
        return _get_non_blank_lines(routecfg)

    if iface == "routes":
        path = _SUSE_NETWORK_ROUTES_FILE
    else:
        path = os.path.join(_SUSE_NETWORK_SCRIPT_DIR, "ifroute-{}".format(iface))

    _write_file_network(routecfg, path)

    return _read_file(path)


def down(iface, iface_type=None):
    """
    Shutdown a network interface

    Args:
        :param iface:
            The name of the interface to shutdown

        :param iface_type:
            The type of the interface
            If ``slave`` is specified, no any action is performing
            Default is ``None``

    Returns:
        str: The result of ``ifdown`` command or ``None`` if ``slave``
        iface_type was specified

    CLI Example:

    .. code-block:: bash

        salt '*' ip.down eth0
    """
    # Slave devices are controlled by the master.
    if not iface_type or iface_type.lower() != "slave":
        return __salt__["cmd.run"]("ifdown {}".format(iface))
    return None


def get_interface(iface):
    """
    Return the contents of an interface script

    Args:
        :param iface:
            The name of the interface to get settings for

    Returns:
        dict: A dictionary of file/content

    CLI Example:

    .. code-block:: bash

        salt '*' ip.get_interface eth0
    """
    path = os.path.join(_SUSE_NETWORK_SCRIPT_DIR, "ifcfg-{}".format(iface))
    return _read_file(path)


def up(iface, iface_type=None):
    """
    Start up a network interface

    Args:
        :param iface:
            The name of the interface to start up

        :param iface_type:
            The type of the interface
            If ``slave`` is specified, no any action is performing
            Default is ``None``

    Returns:
        str: The result of ``ifup`` command or ``None`` if ``slave``
        iface_type was specified

    CLI Example:

    .. code-block:: bash

        salt '*' ip.up eth0
    """
    # Slave devices are controlled by the master.
    if not iface_type or iface_type.lower() != "slave":
        return __salt__["cmd.run"]("ifup {}".format(iface))
    return None


def get_routes(iface):
    """
    Return the contents of the interface routes script.

    Args:
        :param iface:
            The name of the interface to get the routes for

    Returns:
        dict: A dictionary of file/content

    CLI Example:

    .. code-block:: bash

        salt '*' ip.get_routes eth0
    """
    if iface == "routes":
        path = _SUSE_NETWORK_ROUTES_FILE
    else:
        path = os.path.join(_SUSE_NETWORK_SCRIPT_DIR, "ifroute-{}".format(iface))
    return _read_file(path)


def get_network_settings():
    """
    Return the contents of the global network script.

    Args:
        :param iface:
            The name of the interface to start up

        :param iface_type:
            The type of the interface
            If ``slave`` is specified, no any action is performing
            Default is ``None``

    Returns:
        dict: A dictionary of file/content

    CLI Example:

    .. code-block:: bash

        salt '*' ip.get_network_settings
    """
    return _read_file(_SUSE_NETWORK_FILE)


def apply_network_settings(**settings):
    """
    Apply global network configuration.

    Args:
        :param settings:
            The network settings to apply

    Returns:
        The result of ``service.reload`` for ``network`` service

    CLI Example:

    .. code-block:: bash

        salt '*' ip.apply_network_settings
    """
    if "require_reboot" not in settings:
        settings["require_reboot"] = False

    if "apply_hostname" not in settings:
        settings["apply_hostname"] = False

    hostname_res = True
    if settings["apply_hostname"] in _CONFIG_TRUE:
        if "hostname" in settings:
            hostname_res = __salt__["network.mod_hostname"](settings["hostname"])
        else:
            log.warning(
                "The network state sls is trying to apply hostname "
                "changes but no hostname is defined."
            )
            hostname_res = False

    res = True
    if settings["require_reboot"] in _CONFIG_TRUE:
        log.warning(
            "The network state sls is requiring a reboot of the system to "
            "properly apply network configuration."
        )
        res = True
    else:
        res = __salt__["service.reload"]("network")

    return hostname_res and res


def build_network_settings(**settings):
    """
    Build the global network script.

    Args:
        :param settings:
            The network settings

    Returns:
        dict: A dictionary of file/content

    CLI Example:

    .. code-block:: bash

        salt '*' ip.build_network_settings <settings>
    """
    # Read current configuration and store default values
    current_network_settings = _parse_suse_config(_SUSE_NETWORK_FILE)

    # Build settings
    opts = _parse_network_settings(settings, current_network_settings)
    try:
        template = JINJA.get_template("network.jinja")
    except jinja2.exceptions.TemplateNotFound:
        log.error("Could not load template network.jinja")
        return ""
    network = template.render(opts)

    if settings["test"]:
        return _get_non_blank_lines(network)

    # Write settings
    _write_file_network(network, _SUSE_NETWORK_FILE)

    __salt__["cmd.run"]("netconfig update -f")

    return _read_file(_SUSE_NETWORK_FILE)
